r"""
Creating an orthogonal polynomial expansion can be numerical unstable
when using raw statistical moments as input (see `paper by Gautschi`_ for
details). This can be a problem if constructing large expansions since the
error blows up. Given that the distribution is univariate it is instead
possible to create orthogonal polynomials stabilized using the three terms
recurrence relation:

.. math::
    \Phi_{n+1}(q) = \Phi_{n}(q) (q-A_n) - \Phi_{n-1}(q) B_n,

where

.. math::
    A_n = \frac{\left\langle q \Phi_n, \Phi_n \right\rangle}{
          \left\langle \Phi_n, \Phi_n \right\rangle}
        = \frac{\mathbb E[q\Phi_{n}^2]}{
          \mathbb E[\Phi_n^2]}
    B_n = \frac{\left\langle \Phi_n, \Phi_n \right\rangle}{
          \left\langle \Phi_{n-1}, \Phi_{n-1} \right\rangle}
        = \frac{\mathbb E[\Phi_{n}^2]}{\mathbb E[\Phi_{n-1}^2]}

A multivariate polynomial expansion can be created using tensor product
rule of univariate polynomials expansions. This assumes that the
distribution is stochastically independent.

In the ``chaospy`` toolbox three terms recurrence coefficient can be
generating by calling the ``ttr`` instance method::

    >>> dist = chaospy.Uniform(-1, 1)
    >>> dist.ttr([0,1,2,3]).round(4)
    array([[ 0.    ,  0.    ,  0.    ,  0.    ],
           [-0.    ,  0.3333,  0.2667,  0.2571]])

In many of the pre-defined probability distributions in ``chaospy``, the three
terms recurrence coefficients are calculated analytically. If the distribution
does not support the method, the coefficients are instead calculated using the
discretized Stieltjes method (described in the `paper by Golub and Welsch`_).

In ``chaospy`` constructing orthogonal polynomial using the three term
recurrence scheme can be done through ``orth_ttr``. For example::

    >>> dist = chaospy.Iid(chaospy.Gamma(1), 2)
    >>> orths = chaospy.orth_ttr(2, dist)
    >>> orths.round(4)
    polynomial([1.0, q1-1.0, q0-1.0, q1**2-4.0*q1+2.0, q0*q1-q1-q0+1.0,
                q0**2-4.0*q0+2.0])

The method will use the ``ttr`` function if available, and discretized
Stieltjes otherwise.

.. _paper by Gautschi: https://www.ams.org/journals/mcom/1968-22-102/S0025-5718-1968-0228171-0/
.. _paper by Golub and Welsch: https://web.stanford.edu/class/cme335/spr11/S0025-5718-69-99647-1.pdf
"""
import logging

import numpy
import numpoly
import chaospy


def orth_ttr(order, dist, normed=False, graded=True, reverse=True,
             retall=False, cross_truncation=1., sort=None, **kws):
    """
    Create orthogonal polynomial expansion from three terms recurrence formula.

    Args:
        order (int):
            Order of polynomial expansion.
        dist (Distribution):
            Distribution space where polynomials are orthogonal If dist.ttr
            exists, it will be used. Must be stochastically independent.
        normed (bool):
            If True orthonormal polynomials will be used.
        graded (bool):
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**2*q1*q2``,
            ``q0*q1**2*q2`` and ``q0*q1*q2**2``, which all have exponent sum of
            5.
        reverse (bool):
            Reverse lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.
        retall (bool):
            If true return numerical stabilized norms as well. Roughly the same
            as ``cp.E(orth**2, dist)``.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. only include terms where the exponents ``K``
            satisfied the equation
            ``order >= sum(K**(1/cross_truncation))**cross_truncation``.

    Returns:
        (numpoly.ndpoly, numpy.ndarray):
            Orthogonal polynomial expansion. Norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``. Calculated using
            recurrence coefficients for stability.

    Examples:
        >>> distribution = chaospy.J(chaospy.Normal(), chaospy.Normal())
        >>> polynomials, norms = chaospy.orth_ttr(2, distribution, retall=True)
        >>> polynomials.round(10)
        polynomial([1.0, q1, q0, q1**2-1.0, q0*q1, q0**2-1.0])
        >>> norms.round(10)
        array([1., 1., 1., 2., 1., 2.])
        >>> polynomials = chaospy.orth_ttr(2, distribution, normed=True)
        >>> polynomials.round(3)
        polynomial([1.0, q1, q0, 0.707*q1**2-0.707, q0*q1, 0.707*q0**2-0.707])

    """
    logger = logging.getLogger(__name__)
    if sort is not None:
        logger.warning("deprecation warning: 'sort' argument is deprecated; "
                       "use 'graded' and/or 'reverse' instead")
        graded = "G" in sort.upper()
        reverse = "R" not in sort.upper()

    try:
        _, polynomials, norms, = chaospy.quadrature.recurrence.analytical_stieljes(
            numpy.max(order), dist, normed=normed)
    except NotImplementedError:
        abscissas, weights = chaospy.quadrature.generate_quadrature(
            int(10000**(1/len(dist))), dist, rule="fejer")
        _, polynomials, norms, = chaospy.quadrature.recurrence.discretized_stieltjes(
            numpy.max(order), abscissas, weights, normed=normed)

    polynomials = polynomials.reshape((len(dist), numpy.max(order)+1))

    order = numpy.array(order)
    indices = numpoly.glexindex(start=0, stop=order+1, dimensions=len(dist),
                                graded=graded, reverse=reverse,
                                cross_truncation=cross_truncation)
    if len(dist) > 1:
        polynomials = numpoly.prod(chaospy.polynomial([
            poly[idx] for poly, idx in zip(polynomials, indices.T)]), 0)
        norms = numpy.prod([
            norms_[idx] for norms_, idx in zip(norms, indices.T)], 0)
    else:
        polynomials = polynomials.flatten()
        norms = norms.flatten()

    if retall:
        return polynomials, norms
    return polynomials



"""Multivariate Student-T Distribution."""
import numpy
from scipy import special
import chaospy

from .student_t import student_t
from ..baseclass import MeanCovarianceDistribution


class MvStudentT(MeanCovarianceDistribution):
    """
    Multivariate Student-T Distribution.

    Args:
        df (float, Distribution):
            Degree of freedom
        mu (numpy.ndarray, Distribution):
            Location parameter
        sigma (numpy.ndarray):
            Covariance matrix.

    Examples:
        >>> distribution = chaospy.MvStudentT(40, [1, 2], [[1, 0.6], [0.6, 1]])
        >>> distribution
        MvStudentT(df=40, mu=[1, 2], sigma=[[1, 0.6], [0.6, 1]])
        >>> chaospy.Cov(distribution).round(4)
        array([[1.0526, 0.6316],
               [0.6316, 1.0526]])
        >>> mesh = numpy.mgrid[0.25:0.75:3j, 0.25:0.75:2j].reshape(2, -1)
        >>> mesh.round(4)
        array([[0.25, 0.25, 0.5 , 0.5 , 0.75, 0.75],
               [0.25, 0.75, 0.25, 0.75, 0.25, 0.75]])
        >>> inverse_map = distribution.inv(mesh)
        >>> inverse_map.round(4)
        array([[0.3193, 0.3193, 1.    , 1.    , 1.6807, 1.6807],
               [1.0471, 2.1361, 1.4555, 2.5445, 1.8639, 2.9529]])
        >>> numpy.allclose(distribution.fwd(inverse_map), mesh)
        True
        >>> distribution.pdf(inverse_map).round(4)
        array([0.1225, 0.1225, 0.1552, 0.1552, 0.1225, 0.1225])
        >>> distribution.sample(4).round(4)
        array([[ 1.3979, -0.2189,  2.6868,  0.9551],
               [ 3.1625,  0.6234,  1.582 ,  1.7631]])

    """

    def __init__(
            self,
            df,
            mu,
            sigma=None,
            rotation=None,
    ):
        super(MvStudentT, self).__init__(
            dist=student_t(df),
            mean=mu,
            covariance=sigma,
            rotation=rotation,
            repr_args=chaospy.format_repr_kwargs(df=(df, None))+
                      chaospy.format_repr_kwargs(mu=(mu, None))+
                      chaospy.format_repr_kwargs(sigma=(sigma, None)),
        )



"""Independent identical distributed vector of random variables."""
from copy import deepcopy
import numpy
import chaospy

from ..baseclass import Distribution
from .joint import J


class Iid(J):
    """
    Opaque method for creating independent identical distributed random
    variables from an univariate variable.

    Args:
        dist (Distribution):
            Distribution to make into i.i.d. vector. The
            distribution will be copied so to not become a part of
            the dependency graph.
        length (int):
            The length of new distribution.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)
        >>> distribution
        Iid(Normal(mu=0, sigma=1), 2)
        >>> chaospy.Cov(distribution)
        array([[1., 0.],
               [0., 1.]])
        >>> mesh = numpy.mgrid[0.25:0.75:3j, 0.25:0.75:3j].reshape(2, -1)
        >>> mapped_mesh = distribution.inv(mesh)
        >>> mapped_mesh.round(2)
        array([[-0.67, -0.67, -0.67,  0.  ,  0.  ,  0.  ,  0.67,  0.67,  0.67],
               [-0.67,  0.  ,  0.67, -0.67,  0.  ,  0.67, -0.67,  0.  ,  0.67]])
        >>> distribution.fwd(mapped_mesh).round(2)
        array([[0.25, 0.25, 0.25, 0.5 , 0.5 , 0.5 , 0.75, 0.75, 0.75],
               [0.25, 0.5 , 0.75, 0.25, 0.5 , 0.75, 0.25, 0.5 , 0.75]])
        >>> distribution.pdf(mapped_mesh).round(3)
        array([0.101, 0.127, 0.101, 0.127, 0.159, 0.127, 0.101, 0.127, 0.101])
        >>> distribution.sample(4, rule="halton").round(3)
        array([[-1.15 ,  0.319, -0.319,  1.15 ],
               [-0.14 ,  0.765, -0.765,  0.14 ]])
        >>> distribution.mom([[1, 2, 2], [2, 1, 2]]).round(12)
        array([0., 0., 1.])

    """

    def __init__(self, dist, length, rotation=None):
        assert isinstance(dist, Distribution)
        assert len(dist) == 1 and length >= 1
        assert len(dist._dependencies[0]) == 1
        exclusion = dist._dependencies[0].copy()
        dists = [deepcopy(dist) for _ in range(length)]
        for dist in dists:
            dist._dependencies = chaospy.init_dependencies(
                dist, rotation=[0], dependency_type="iid")
        super(Iid, self).__init__(*dists, rotation=rotation)
        self._exclusion.update(exclusion)
        self._repr_args = [dist, length]



"""Generalized exponential distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class generalized_exponential(SimpleDistribution):
    """Generalized exponential distribution."""

    def __init__(self, a=1, b=1, c=1):
        super(generalized_exponential, self).__init__(dict(a=a, b=b, c=c))

    def _pdf(self, x, a, b, c):
        return (a+b*(-numpy.expm1(-c*x)))*numpy.exp((-a-b)*x+b*(-numpy.expm1(-c*x))/c)

    def _cdf(self, x, a, b, c):
        output = -numpy.expm1((-a-b)*x + b*(-numpy.expm1(-c*x))/c)
        output = numpy.where(x > 0, output, 0)
        return output

    def _lower(self, a, b, c):
        return 0.

    def _upper(self, a, b, c):
        qloc, a, b, c = numpy.broadcast_arrays(1-1e-12, a, b, c)
        return chaospy.approximate_inverse(
            distribution=self,
            idx=0,
            qloc=qloc,
            parameters=dict(a=a, b=b, c=c),
            bounds=(0., 500./a/b/c),
            tolerance=1e-15,
        )


class GeneralizedExponential(ShiftScaleDistribution):
    """
    Generalized exponential distribution.

    Args:
        a (float, Distribution):
            First shape parameter
        b (float, Distribution):
            Second shape parameter
        c (float, Distribution):
            Third shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Note:
        "An Extension of Marshall and Olkin's Bivariate Exponential Distribution",
        H.K. Ryu, Journal of the American Statistical Association, 1993.

        "The Exponential Distribution: Theory, Methods and Applications",
        N. Balakrishnan, Asit P. Basu.

    Examples:
        >>> distribution = chaospy.GeneralizedExponential(3, 4, 5)
        >>> distribution
        GeneralizedExponential(3, 4, 5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.063, 0.127, 0.204, 0.321, 4.062])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([3.   , 3.26 , 2.925, 2.223, 1.24 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.006, 0.108, 0.089, 0.04 ])

    """

    def __init__(self, a=1, b=1, c=1, scale=1, shift=0):
        super(GeneralizedExponential, self).__init__(
            dist=generalized_exponential(a, b, c),
            scale=scale,
            shift=shift,
            repr_args=[a, b, c],
        )



"""(Non-central) F distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class f(SimpleDistribution):
    """F distribution."""

    def __init__(self, dfn, dfd, nc):
        super(f, self).__init__(
            parameters=dict(dfn=dfn, dfd=dfd, nc=nc),
            repr_args=[dfn, dfd, "nc=%s" % nc],
        )

    def _pdf(self, x, dfn, dfd, nc):
        n1, n2 = dfn, dfd
        term = -nc/2.+nc*n1*x/(2*(n2+n1*x)) + special.gammaln(n1/2.)+special.gammaln(1+n2/2.)
        term -= special.gammaln((n1+n2)/2.)
        Px = numpy.exp(term)
        Px *= n1**(n1/2.) * n2**(n2/2.) * x**(n1/2.-1)
        Px *= (n2+n1*x)**(-(n1+n2)/2.)
        Px *= special.assoc_laguerre(-nc*n1*x/(2.*(n2+n1*x)), n2/2., n1/2.-1)
        Px /= special.beta(n1/2., n2/2.)
        return Px

    def _cdf(self, x, dfn, dfd, nc):
        return special.ncfdtr(dfn, dfd, nc, x)

    def _ppf(self, q, dfn, dfd, nc):
        return numpy.where(q == 1, self._upper(dfn, dfd, nc), special.ncfdtri(dfn, dfd, nc, q))

    def _lower(self, dfn, dfd, nc):
        return 0.

    def _upper(self, dfn, dfd, nc):
        return special.ncfdtri(dfn, dfd, nc, 1-1e-10)


class F(ShiftScaleDistribution):
    """
    (Non-central) F or Fisher-Snedecor distribution.

    Args:
        n (float, Distribution):
            Degres of freedom for numerator
        m (float, Distribution):
            Degres of freedom for denominator
        nc (float, Distribution):
            Non-centrality parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.F(10, 10, 0)
        >>> distribution
        F(10, 10, nc=0)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([  0.   ,   0.578,   0.848,   1.179,   1.732, 261.403])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.734, 0.701, 0.505, 0.245, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.292, 0.455, 2.984, 0.971])

    """

    def __init__(self, n=2, m=10, nc=0, shift=0, scale=1):
        super(F, self).__init__(
            dist=f(n, m, nc),
            shift=shift,
            scale=scale,
            repr_args=[n, m, "nc=%s" % nc],
        )



"""Expected value."""
import numpy
import numpoly


def E(poly, dist=None, **kws):
    """
    Expected value operator.

    1st order statistics of a probability distribution or polynomial on a given
    probability space.

    Args:
        poly (numpoly.ndpoly, Distribution):
            Input to take expected value on.
        dist (Distribution):
            Defines the space the expected value is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            The expected value of the polynomial or distribution, where
            ``expected.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.E(dist)
        array([1., 0.])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.E(poly, dist)
        array([ 1.,  1.,  0., -1.])

    """
    if dist is None:
        dist, poly = poly, numpoly.variable(len(poly))

    poly = numpoly.set_dimensions(poly, len(dist))
    if poly.isconstant():
        return poly.tonumpy()

    moments = dist.mom(poly.exponents.T, **kws)
    if len(dist) == 1:
        moments = moments[0]

    out = numpy.zeros(poly.shape)
    for idx, key in enumerate(poly.keys):
        out += poly[key]*moments[idx]
    return out



"""
``chaospy`` is in no means a package that provides absolutely all functionality
for all problems. If another project provides functionality that can work well
in tandem, the best approach, if possible, is to use both at the same time. To
make such an approach more feasible, some compatibility wrappers exists
allowing for using components from other projects as part of ``chaospy``.
"""
from .openturns_ import openturns_dist, OpenTURNSDist
from .scipy_stats import ScipyStatsDist
from .samples import sample_dist, SampleDist



"""Log-Normal probability distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class log_normal(SimpleDistribution):

    def __init__(self, a=1):
        super(log_normal, self).__init__(dict(a=a))

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return numpy.e**(a*6.37)

    def _pdf(self, x, a):
        out = (numpy.e**(-numpy.log(x+(1-x)*(x<=0))**2/(2*a*a))/
               ((x+(1-x)*(x<=0))*a*numpy.sqrt(2*numpy.pi))*(x>0))
        return out

    def _cdf(self, x, a):
        return special.ndtr(numpy.log(x+(1-x)*(x<=0))/a)*(x>0)

    def _ppf(self, x, a):
        return numpy.e**(a*special.ndtri(x))

    def _mom(self, k, a):
        return numpy.e**(.5*a*a*k*k)

    def _ttr(self, n, a):
        """Stieltjes-Wigert."""
        return (
            (numpy.e**(n*a*a)*(numpy.e**(a*a)+1)-1)*numpy.e**(.5*(2*n-1)*a*a),
            (numpy.e**(n*a*a)-1)*numpy.e**((3*n-2)*a*a)
        )


class LogNormal(ShiftScaleDistribution):
    R"""
    Log-normal distribution

    Args:
        mu (float, Distribution):
            Mean in the normal distribution.  Overlaps with scale by
            mu=log(scale)
        sigma (float, Distribution):
            Standard deviation of the normal distribution.
        shift (float, Distribution):
            Location of the lower bound.
        scale (float, Distribution):
            Scale parameter. Overlaps with mu by scale=e**mu

    Examples:
        >>> distribution = chaospy.LogNormal(0, 0.1)
        >>> distribution
        LogNormal(mu=0, sigma=0.1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.919, 0.975, 1.026, 1.088, 1.891])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 3.045, 3.963, 3.767, 2.574, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.04 , 0.887, 1.179, 0.996])
        >>> distribution.mom(1).round(3)
        1.005
        >>> distribution.ttr([0, 1, 2, 3]).round(3)
        array([[1.005, 1.035, 1.067, 1.098],
               [0.   , 0.01 , 0.021, 0.033]])

    """

    def __init__(self, mu=0, sigma=1, shift=0, scale=1):
        dist = ShiftScaleDistribution(dist=log_normal(sigma), scale=numpy.e**mu)
        super(LogNormal, self).__init__(
            dist=dist,
            scale=scale,
            shift=shift,
            repr_args=["mu=%s" % mu, "sigma=%s" % sigma],
        )


class Gilbrat(ShiftScaleDistribution):
    """
    Gilbrat distribution.

    Standard log-normal distribution

    Args:
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Gilbrat(scale=0.0015)
        >>> distribution
        Gilbrat(scale=0.0015)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.001, 0.001, 0.002, 0.003, 0.876])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([  0.   , 433.031, 331.825, 199.919,  80.444,   0.   ])
        >>> distribution.sample(4).round(4)
        array([0.0022, 0.0005, 0.0078, 0.0014])
        >>> distribution.mom(1).round(8)
        0.00247308
        >>> distribution.ttr([0, 1, 2]).round(4)
        array([[0.0025, 0.0225, 0.178 ],
               [0.    , 0.    , 0.0008]])

    """

    def __init__(self, scale=1, shift=0):
        super(Gilbrat, self).__init__(
            dist=log_normal(1),
            scale=scale,
            shift=shift,
            repr_args=[],
        )



"""Power normal or Box-Cox distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class power_normal(SimpleDistribution):
    """Power normal or Box-Cox distribution."""

    def __init__(self, c):
        super(power_normal, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        norm = (2*numpy.pi)**-.5*numpy.exp(-x**2/2.)
        return c*norm*special.ndtr(-x)**(c-1.)

    def _cdf(self, x, c):
        return 1.-special.ndtr(-x)**c

    def _ppf(self, q, c):
        return -special.ndtri(pow(1-q, 1./c))

    def _lower(self, c):
        return -special.ndtri(pow(1-1e-15, 1./c))

    def _upper(self, c):
        return -special.ndtri(pow(1e-15, 1./c))


class PowerNormal(ShiftScaleDistribution):
    """
    Power normal or Box-Cox distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        mu (float, Distribution):
            Mean of the normal distribution
        scale (float, Distribution):
            Standard deviation of the normal distribution

    Examples:
        >>> distribution = chaospy.PowerNormal(1)
        >>> distribution
        PowerNormal(1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-7.941, -0.842, -0.253,  0.253,  0.842,  7.941])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.28 , 0.386, 0.386, 0.28 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.395, -1.2  ,  1.648, -0.045])

    """

    def __init__(self, shape=1, mu=0, sigma=1):
        super(PowerNormal, self).__init__(
            dist=power_normal(shape),
            scale=sigma,
            shift=mu,
        )
        self._repr_args = [shape]
        self._repr_args += chaospy.format_repr_kwargs(mu=(mu, 0))
        self._repr_args += chaospy.format_repr_kwargs(sigma=(sigma, 1))



"""Spearman's correlation coefficient."""
from scipy.stats import spearmanr


def Spearman(poly, dist, sample=10000, retall=False, **kws):
    """
    Calculate Spearman's rank-order correlation coefficient.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest.
        dist (Distribution):
            Defines the space where correlation is taken.
        sample (int):
            Number of samples used in estimation.
        retall (bool):
            If true, return p-value as well.

    Returns:
        (float, numpy.ndarray):
            Correlation output ``rho``. Of type float if two-dimensional problem.
            Correleation matrix if larger.
        (float, numpy.ndarray):
            The two-sided p-value for a hypothesis test whose null hypothesis
            is that two sets of data are uncorrelated, has same dimension as
            ``rho``.
    """
    samples = dist.sample(sample, **kws)
    poly = polynomials.flatten(poly)
    Y = poly(*samples)
    if retall:
        return spearmanr(Y.T)
    return spearmanr(Y.T)[0]



"""Multivariate Log-Normal Distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import Distribution


class MvLogNormal(Distribution):
    """
    Multivariate Log-Normal Distribution.

    Args:
        loc (float, Distribution):
            Mean vector
        scale (float, Distribution):
            Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvLogNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> distribution
        MvLogNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> distribution.inv(mesh).round(4)
        array([[[ 1.3847,  2.7183,  5.3361],
                [ 1.3847,  2.7183,  5.3361],
                [ 1.3847,  2.7183,  5.3361]],
        <BLANKLINE>
               [[ 2.874 ,  4.3077,  6.4566],
                [ 4.9298,  7.3891, 11.075 ],
                [ 8.4562, 12.6745, 18.9971]]])
        >>> distribution.fwd(distribution.inv(mesh)).round(4)
        array([[[0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75]],
        <BLANKLINE>
               [[0.25, 0.25, 0.25],
                [0.5 , 0.5 , 0.5 ],
                [0.75, 0.75, 0.75]]])
        >>> distribution.pdf(distribution.inv(mesh)).round(4)
        array([[0.0108, 0.002 , 0.    ],
               [0.0135, 0.0035, 0.    ],
               [0.0107, 0.0038, 0.0001]])
        >>> distribution.sample(4).round(4)
        array([[ 4.0351,  0.8185, 14.1201,  2.5996],
               [23.279 ,  1.8986,  4.9261,  5.8399]])
        >>> distribution.mom((1, 2)).round(4)
        6002.9122
    """

    def __init__(
            self,
            mu,
            sigma,
            rotation=None,
    ):
        repr_args = [mu, sigma]
        dependencies, parameters, rotation = chaospy.declare_dependencies(
            self,
            parameters=dict(mu=mu, sigma=sigma),
            rotation=rotation,
            dependency_type="accumulate",
        )

        sigma = parameters["sigma"]
        assert sigma.ndim == 2, (
            "Covariance must either be scalar, vector or matrix")

        self._permute = numpy.eye(len(rotation), dtype=int)[rotation]
        self._psigma = self._permute.dot(sigma).dot(self._permute.T)
        cholesky = numpy.linalg.cholesky(self._psigma)
        self._fwd_transform = self._permute.T.dot(numpy.linalg.inv(cholesky))
        self._inv_transform = self._permute.T.dot(cholesky)

        super(MvLogNormal, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            rotation=rotation,
            repr_args=repr_args,
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(MvLogNormal, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        if idx is None:
            parameters.pop("idx")
        return parameters

    def _pdf(self, xloc, idx, mu, sigma, cache):
        dim = self._rotation.index(idx)
        if dim:
            covinv = numpy.linalg.inv(self._psigma[:dim, :dim])
            mu_transform = self._psigma[dim, :dim].dot(covinv)
            assert numpy.isfinite(sigma).all(), (dim, self._psigma)
            conditions = [self._get_cache(dim_, cache, get=0)
                          for dim_ in self._rotation[:dim]]
            assert not any([isinstance(condition, chaospy.Distribution)
                            for condition in conditions])
            mu = mu[numpy.asarray(self._rotation[:dim+1])]
            mu = mu[dim]+mu_transform.dot((numpy.vstack(conditions).T-mu[:dim]).T)
            sigma = numpy.sqrt(
                self._psigma[dim, dim]-
                self._psigma[dim, :dim].dot(covinv).dot(self._psigma[:dim, dim])
            )
        else:
            mu = mu[dim]
            sigma = numpy.sqrt(self._psigma[0, 0])

        out = numpy.zeros(xloc.shape)
        indices = xloc > 0
        zloc = (numpy.log(xloc[indices])-mu)/sigma
        out[indices] = 1/numpy.sqrt(2*numpy.pi)*numpy.exp(-zloc**2/2.)/xloc[indices]
        return out

    def _cdf(self, xloc, idx, mu, sigma, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        yloc = numpy.vstack(conditions+[xloc])
        yloc = numpy.log(numpy.abs(yloc) + numpy.where(yloc <= 0, 1., 0.))
        mu = mu[numpy.asarray(self._rotation[:len(yloc)])]
        loc = (yloc.T-mu).T
        zloc = self._fwd_transform[idx, :len(yloc)].dot(loc)
        out = special.ndtr(zloc)
        return numpy.where(xloc <= 0, 0., out)

    def _ppf(self, uloc, idx, mu, sigma, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=1)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        uloc = numpy.vstack(conditions+[uloc])
        zloc = special.ndtri(uloc)
        loc = self._inv_transform[idx, :len(uloc)].dot(zloc)
        xloc = loc+mu[idx]
        out = numpy.e**xloc
        return out

    def _mom(self, kloc, mu, sigma, cache):
        output =  numpy.dot(kloc, mu)
        output += .5*numpy.dot(numpy.dot(kloc, sigma), kloc)
        output =  numpy.e**(output)
        return output

    def _lower(self, idx, mu, sigma, cache):
        return 0.

    def _upper(self, idx, mu, sigma, cache):
        return numpy.exp(7.1*numpy.sqrt(sigma[idx, idx])) + mu[idx]



"""Multivariate Normal Distribution."""
import logging
import numpy
from scipy import special
import chaospy

from .normal import normal
from ..baseclass import MeanCovarianceDistribution


class MvNormal(MeanCovarianceDistribution):
    r"""
    Multivariate Normal Distribution.

    Args:
        mu (float, numpy.ndarray):
            Mean vector
        scale (float, numpy.ndarray):
            Covariance matrix or variance vector if scale is a 1-d vector.
        rotation (Sequence[int]):
            The order of which to resolve conditionals.
            Defaults to `range(len(distribution))` which is the same as
            `p(x0), p(x1|x0), p(x2|x0,x1), ...`.

    Examples:
        >>> distribution = chaospy.MvNormal([10, 20, 30],
        ...     [[1, 0.2, 0.3], [0.2, 2, 0.4], [0.3, 0.4, 1]], rotation=[1, 2, 0])
        >>> distribution  # doctest: +NORMALIZE_WHITESPACE
        MvNormal(mu=[10, 20, 30],
                 sigma=[[1, 0.2, 0.3], [0.2, 2, 0.4], [0.3, 0.4, 1]])
        >>> chaospy.E(distribution)
        array([10., 20., 30.])
        >>> chaospy.Cov(distribution)
        array([[1. , 0.2, 0.3],
               [0.2, 2. , 0.4],
               [0.3, 0.4, 1. ]])
        >>> mesh = numpy.mgrid[:2, :2, :2].reshape(3, -1)*.5+.1
        >>> mesh
        array([[0.1, 0.1, 0.1, 0.1, 0.6, 0.6, 0.6, 0.6],
               [0.1, 0.1, 0.6, 0.6, 0.1, 0.1, 0.6, 0.6],
               [0.1, 0.6, 0.1, 0.6, 0.1, 0.6, 0.1, 0.6]])
        >>> mapped_samples = distribution.inv(mesh)
        >>> mapped_samples.round(2)
        array([[ 8.25,  8.67,  8.47,  8.88,  9.71, 10.13,  9.93, 10.35],
               [18.19, 18.19, 20.36, 20.36, 18.19, 18.19, 20.36, 20.36],
               [28.41, 29.88, 28.84, 30.31, 28.41, 29.88, 28.84, 30.31]])
        >>> numpy.allclose(distribution.fwd(mapped_samples), mesh)
        True
        >>> distribution.pdf(mapped_samples).round(4)
        array([0.0042, 0.0092, 0.0092, 0.0203, 0.0092, 0.0203, 0.0203, 0.0446])
        >>> distribution.sample(4).round(4)
        array([[10.3396,  9.0158, 11.1009, 10.0971],
               [21.6096, 18.871 , 17.5357, 19.6314],
               [29.6231, 30.7349, 28.7239, 30.5507]])

    """

    def __init__(
            self,
            mu,
            sigma=None,
            rotation=None,
    ):
        super(MvNormal, self).__init__(
            dist=normal(),
            mean=mu,
            covariance=sigma,
            rotation=rotation,
            repr_args=chaospy.format_repr_kwargs(mu=(mu, None))+
                      chaospy.format_repr_kwargs(sigma=(sigma, None)),
        )

    def _mom(self, k, mean, sigma, cache):
        if isinstance(mean, chaospy.Distribution):
            mean = mean._get_cache(None, cache=cache, get=0)
            if isinstance(mean, chaospy.Distribution):
                raise chaospy.UnsupportedFeature(
                    "Analytical moment of a conditional not supported")
        out = 0.
        for idx, kdx in enumerate(numpy.ndindex(*[_+1 for _ in k])):
            coef = numpy.prod(special.comb(k.T, kdx).T, 0)
            diff = k.T-kdx
            pos = diff >= 0
            diff = diff*pos
            pos = numpy.all(pos)
            location_ = numpy.prod(mean**diff, -1)

            out = out+pos*coef*location_*isserlis_moment(tuple(kdx), sigma)

        return out


def isserlis_moment(k, scale):
    """
    Centralized statistical moments using Isserlis' theorem.

    Args:
        k (Tuple[int, ...]):
            Moment orders.
        scale (ndarray):
            Covariance matrix defining dependencies between variables.

    Returns:
        Raw statistical moment of order ``k`` given covariance ``scale``.

    Examples:
        >>> scale = 0.5*numpy.eye(3)+0.5
        >>> isserlis_moment((2, 2, 2), scale)
        3.5
        >>> isserlis_moment((0, 0, 0), scale)
        1.0
        >>> isserlis_moment((1, 0, 0), scale)
        0.0
        >>> isserlis_moment((0, 1, 1), scale)
        0.5
        >>> isserlis_moment((0, 0, 2), scale)
        1.0
    """
    if scale.ndim == 2:
        scale = scale[numpy.newaxis]
        return isserlis_moment(k, scale)[0]

    if not isinstance(k, numpy.ndarray):
        k = numpy.asarray(k)
    assert len(k.shape) == 1

    # Recursive exit condition
    if (numpy.sum(k) % 2 == 1) or numpy.any(k < 0):
        return numpy.zeros(len(scale))

    # Choose a pivot index as first non-zero entry
    idx = numpy.nonzero(k)[0]
    if not idx.size:
        # Skip if no pivot found
        return numpy.ones(len(scale))
    idx = idx[0]

    eye = numpy.eye(len(k), dtype=int)
    out = (k[idx]-1)*scale[:, idx, idx]*isserlis_moment(k-2*eye[idx], scale)
    for idj in range(idx+1, len(k)):
        out += k[idj]*scale[:, idx, idj]*isserlis_moment(k-eye[idx]-eye[idj], scale)
    return out



"""Power operator."""
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Power(OperatorDistribution):
    """Power operator."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Distribution, numpy.ndarray) : Left hand side.
            right (Distribution, numpy.ndarray) : Right hand side.
        """
        super(Power, self).__init__(
            left=left,
            right=right,
            repr_args=[left, right],
        )

    def _lower(self, idx, left, right, cache):
        """
        Distribution lower bounds.

        Example:
            >>> chaospy.Uniform().lower
            array([0.])
            >>> chaospy.Power(chaospy.Uniform(), 2).lower
            array([0.])
            >>> chaospy.Power(chaospy.Uniform(1, 2), -1).lower
            array([0.5])
            >>> chaospy.Power(2, chaospy.Uniform()).lower
            array([1.])
            >>> chaospy.Power(2, chaospy.Uniform(-1, 0)).lower
            array([0.5])

        """
        left = self._parameters["left"]
        right = self._parameters["right"]

        if isinstance(left, Distribution):
            left_upper = left._get_upper(idx, cache=self._upper_cache)
            left_lower = left._get_lower(idx, cache=self._lower_cache)

            if isinstance(right, Distribution):
                right_upper = right._get_upper(idx, cache=self._upper_cache)
                right_lower = right._get_lower(idx, cache=self._lower_cache)

                out = numpy.min(numpy.broadcast_arrays(
                    left_lower.T**right_lower.T,
                    left_lower.T**right_upper.T,
                    left_upper.T**right_lower.T,
                    left_upper.T**right_upper.T,
                ), axis=0).T

            else:
                # assert 0, (idx, left_lower, left_upper, right)
                out = numpy.min([left_lower**right[idx],
                                 left_upper**right[idx]], axis=0).T

        else:
            assert isinstance(right, Distribution)
            right_upper = right._get_upper(idx, cache=self._upper_cache)
            right_lower = right._get_lower(idx, cache=self._lower_cache)
            out = numpy.min([left[idx]**right_lower,
                             left[idx]**right_upper], axis=0).T

        return out

    def _upper(self, idx, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> chaospy.Uniform().upper
            array([1.])
            >>> chaospy.Power(chaospy.Uniform(), 2).upper
            array([1.])
            >>> chaospy.Power(chaospy.Uniform(1, 2), -1).upper
            array([1.])
            >>> chaospy.Power(2, chaospy.Uniform()).upper
            array([2.])
            >>> chaospy.Power(2, chaospy.Uniform(-1, 0)).upper
            array([1.])

        """
        left = self._parameters["left"]
        right = self._parameters["right"]
        if isinstance(left, Distribution):
            left_lower = left._get_lower(idx, cache=self._lower_cache)
            left_upper = left._get_upper(idx, cache=self._upper_cache)

            if isinstance(right, Distribution):
                right_lower = right._get_lower(idx, cache=self._lower_cache)
                right_upper = right._get_upper(idx, cache=self._upper_cache)

                out = numpy.max(numpy.broadcast_arrays(
                    (left_lower.T**right_lower.T).T,
                    (left_lower.T**right_upper.T).T,
                    (left_upper.T**right_lower.T).T,
                    (left_upper.T**right_upper.T).T,
                ), axis=0)

            else:
                out = numpy.max([left_lower**right[idx],
                                 left_upper**right[idx]], axis=0)

        else:
            assert isinstance(right, Distribution)
            right_lower = right._get_lower(idx, cache=self._lower_cache)
            right_upper = right._get_upper(idx, cache=self._upper_cache)
            out = numpy.max([left[idx]**right_lower,
                             left[idx]**right_upper], axis=0)

        return out

    def _pdf(self, xloc, idx, left, right, cache):
        """
        Probability density function.

        Example:
            >>> chaospy.Uniform().pdf([-0.5, 0.5, 1.5, 2.5])
            array([0., 1., 0., 0.])
            >>> chaospy.Power(chaospy.Uniform(), 2).pdf([-0.5, 0.5, 1.5, 2.5])
            array([0.        , 0.70710678, 0.        , 0.        ])
            >>> chaospy.Power(chaospy.Uniform(1, 2), -1).pdf([0.4, 0.6, 0.8, 1.2])
            array([0.        , 2.77777778, 1.5625    , 0.        ])
            >>> chaospy.Power(2, chaospy.Uniform()).pdf([-0.5, 0.5, 1.5, 2.5])
            array([0.        , 0.        , 0.96179669, 0.        ])
            >>> chaospy.Power(2, chaospy.Uniform(-1, 0)).pdf([0.4, 0.6, 0.8, 1.2])
            array([0.        , 2.40449173, 1.8033688 , 0.        ])

        """
        if isinstance(left, Distribution):
            x_ = numpy.sign(xloc)*numpy.abs(xloc)**(1./right-1)
            xloc = numpy.sign(xloc)*numpy.abs(xloc)**(1./right)
            pairs = numpy.sign(xloc**right) == 1
            out = left._get_pdf(xloc, idx, cache=cache.copy())
            if numpy.any(pairs):
                out = out+pairs*left._get_pdf(-xloc, idx, cache=cache)
            out = numpy.sign(right)*out*x_/right
            out[numpy.isnan(out)] = numpy.inf

        else:
            assert numpy.all(left > 0), "imaginary result"
            x_ = numpy.where(xloc <= 0, -numpy.inf,
                             numpy.log(xloc + 1.*(xloc<=0))/numpy.log(left+1.*(left == 1)))
            num_ = numpy.log(left+1.*(left == 1))*xloc
            num_ = num_ + 1.*(num_==0)
            out = right._get_pdf(x_, idx, cache=cache)/num_

        return out

    def _cdf(self, xloc, idx, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> chaospy.Uniform().fwd([-0.5, 0.5, 1.5, 2.5])
            array([0. , 0.5, 1. , 1. ])
            >>> chaospy.Power(chaospy.Uniform(), 2).fwd([-0.5, 0.5, 1.5, 2.5])
            array([0.        , 0.70710678, 1.        , 1.        ])
            >>> chaospy.Power(chaospy.Uniform(1, 2), -1).fwd([0.4, 0.6, 0.8, 1.2])
            array([0.        , 0.33333333, 0.75      , 1.        ])
            >>> chaospy.Power(2, chaospy.Uniform()).fwd([-0.5, 0.5, 1.5, 2.5])
            array([0.       , 0.       , 0.5849625, 1.       ])
            >>> chaospy.Power(2, chaospy.Uniform(-1, 0)).fwd([0.4, 0.6, 0.8, 1.2])
            array([0.        , 0.26303441, 0.67807191, 1.        ])

        """
        if isinstance(left, Distribution):
            y = numpy.sign(xloc)*numpy.abs(xloc)**(1./right)
            pairs = numpy.sign(xloc**right) != -1
            out2 = left._get_fwd(-y, idx, cache=cache.copy())
            out1 = left._get_fwd(y, idx, cache=cache)
            out = numpy.where(right < 0, 1-out1, out1-pairs*out2)
        else:
            y = (numpy.log(numpy.abs(xloc)+1.*(xloc <= 0))/
                 numpy.log(numpy.abs(left)+1.*(left == 1)))
            out = right._get_fwd(y, idx, cache=cache)
            out = numpy.where(xloc <= 0, 0., out)
        return out

    def _ppf(self, q, idx, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> chaospy.Uniform().inv([0.1, 0.2, 0.9])
            array([0.1, 0.2, 0.9])
            >>> chaospy.Power(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9])
            array([0.01, 0.04, 0.81])
            >>> chaospy.Power(chaospy.Uniform(1, 2), -1).inv([0.1, 0.2, 0.9])
            array([0.52631579, 0.55555556, 0.90909091])
            >>> chaospy.Power(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9])
            array([1.07177346, 1.14869835, 1.86606598])
            >>> chaospy.Power(2, chaospy.Uniform(-1, 0)).inv([0.1, 0.2, 0.9])
            array([0.53588673, 0.57434918, 0.93303299])

        """
        if isinstance(left, Distribution):
            q = numpy.where(right.T < 0, 1-q.T, q.T).T
            out = (left._get_inv(q, idx, cache=cache).T**right.T).T
        else:
            out = right._get_inv(q, idx, cache=cache)
            out = numpy.where(left < 0, 1-out, out)
            out = (left.T**out.T).T
        return out

    def _mom(self, k, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4)
            array([1.    , 0.5   , 0.3333, 0.25  ])
            >>> numpy.around(chaospy.Power(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4)
            array([1.    , 0.3333, 0.2   , 0.1429])
            >>> numpy.around(chaospy.Power(chaospy.Uniform(1, 2), -1).mom([0, 1, 2, 3]), 4)
            array([1.    , 0.6931, 0.5   , 0.375 ])
            >>> numpy.around(chaospy.Power(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4)
            array([1.    , 1.4427, 2.164 , 3.3663])
            >>> numpy.around(chaospy.Power(2, chaospy.Uniform(-1, 0)).mom([0, 1, 2, 3]), 4)
            array([1.    , 0.7213, 0.541 , 0.4208])

        """
        del cache
        if isinstance(right, Distribution):
            raise chaospy.UnsupportedFeature(
                "distribution as exponent not supported.")
        if numpy.any(right < 0):
            raise chaospy.UnsupportedFeature(
                "distribution to negative power not supported.")
        if not numpy.allclose(right, numpy.array(right, dtype=int)):
            raise chaospy.UnsupportedFeature(
                "distribution to fractional power not supported.")
        return left._get_mom(k*right)



"""
Create samples from the `Halton sequence`_.

In statistics, Halton sequences are sequences used to generate points in space
for numerical methods such as Monte Carlo simulations. Although these sequences
are deterministic, they are of low discrepancy, that is, appear to be random
for many purposes. They were first introduced in 1960 and are an example of
a quasi-random number sequence. They generalise the one-dimensional van der
Corput sequences.

Example usage
-------------

Standard usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(3, rule="halton")
    >>> samples.round(4)
    array([[0.125 , 0.625 , 0.375 ],
           [0.4444, 0.7778, 0.2222]])
    >>> samples = distribution.sample(4, rule="halton")
    >>> samples.round(4)
    array([[0.125 , 0.625 , 0.375 , 0.875 ],
           [0.4444, 0.7778, 0.2222, 0.5556]])

.. _Halton sequence: https://en.wikipedia.org/wiki/Halton_sequence
"""
import numpy

from .van_der_corput import create_van_der_corput_samples
from .primes import create_primes


def create_halton_samples(order, dim=1, burnin=-1, primes=()):
    """
    Create Halton sequence.

    For ``dim == 1`` the sequence falls back to Van Der Corput sequence.

    Args:
        order (int):
            The order of the Halton sequence. Defines the number of samples.
        dim (int):
            The number of dimensions in the Halton sequence.
        burnin (int):
            Skip the first ``burnin`` samples. If negative, the maximum of
            ``primes`` is used.
        primes (tuple):
            The (non-)prime base to calculate values along each axis. If
            empty, growing prime values starting from 2 will be used.

    Returns (numpy.ndarray):
        Halton sequence with ``shape == (dim, order)``.
    """
    primes = list(primes)
    if not primes:
        prime_order = 10*dim
        while len(primes) < dim:
            primes = create_primes(prime_order)
            prime_order *= 2
    primes = primes[:dim]
    assert len(primes) == dim, "not enough primes"

    if burnin < 0:
        burnin = max(primes)

    out = numpy.empty((dim, order))
    indices = [idx+burnin for idx in range(order)]
    for dim_ in range(dim):
        out[dim_] = create_van_der_corput_samples(
            indices, number_base=primes[dim_])
    return out



r"""
Joe Copula.

The calculation of the derivative of the ``iphi`` function:

.. math::
  \begin{eqnarray}
    iphi(u)     &= 1-(1-e^{-u})^{1/\theta} \\
    iphi'(u)    &= -(1/theta)*(1-e^(-u))^(1/theta-1)*e^(-u)
                = -\sigma(1-e^-u, 1, \theta) = j1 \\
    iphi''(u)   &= d/du ( j1 )
                = j2 - j1
                = j2 - iphi'(u) \\
    iphi'''(u)  &= d/du ( j2 - j1 )
                = j3 - 2j2 - j2 + j1
                = (j3 - 2j2) - (j2 - j1)
                = (j3 - 2j2) - iphi''(u) \\
    iphi''''(u) &= d/du ( (j3 - 2j2) - (j2 - j1) )
                = ((j4 - 3j3) - 2(j3-2j2)) - ((j3-2j2) - (j2-j1))
                = ((j4 - 3j3) - 2(j3-2j2)) - iphi'''(u) \\
    iphi'''''(u)&= d/du ((j4 - 3j3) - 2(j3-2j2)) - ((j3-2j2) - (j2-j1))
                = ((j5-4j4) - 3(j4-3j3)) - 2((j4-3j3)-2(j3-2j2)) -
                   (((j4-3j3)-2(j3-2j2)) - ((j3-2j2)-(j2-j1)))
                = ((j5-4j4) - 3(j4-3j3)) - 2((j4-3j3)-2(j3-2j2)) - iphi''''(u) \\
  \end{eqnarray}

Here the notation :math:`jn` is a short hand and means:

.. math::
    jn = J(u, n) = -\sigma(1-e^-u, n) e^{-un}

which was the property:

.. math::

    d/du j(u, n) = J(u, n+1) - n*J(u, n)

This problem can be solved recursively using the function:

.. math::
  \begin{eqnarray}
    \rho(u, n, \theta, m) &= \sigma(1-e^{-u}, \theta, n) e^{-n\theta} & n &= m \\
    \rho(u, n, \theta, m) &= \rho(u, n, \theta, m+1) - m \rho(u, n-1, \theta, m) & n &\neq m \\
  \end{eqnarray}

Solution is then just:
.. math::
    iphi^{(n)}(u) = \rho(u, n, \theta, 1)
"""
try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache
import numpy
from scipy import special

from ..baseclass import CopulaDistribution
from .archimedean import Archimedean


class joe(Archimedean):
    """Joe copula."""

    def _phi(self, t_loc, theta):
        return -numpy.log(1-(1-t_loc)**theta)

    def _delta_phi(self, t_loc, theta):
        return theta*(1-t_loc)**(theta-1)/(1-(1-t_loc)**theta)

    def _inverse_phi(self, u_loc, theta, order):
        if not order:
            return 1-(1-numpy.e**-u_loc)**(1/theta)
        @lru_cache(None)
        def rho(n, m=1):
            if n == m:
                return self._sigma(1-numpy.e**-u_loc, theta, n)*numpy.e**(-n*theta)
            return rho(n, m+1)-m*rho(n-1, m)
        return rho(order)


class Joe(CopulaDistribution):
    """
    Joe Copula

    Args:
        dist (Distribution):
            The distribution to wrap
        theta (float):
            Copula parameter. Required to be above 1.

    Examples:
        >>> distribution = chaospy.Joe(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2), theta=2)
        >>> distribution
        Joe(Iid(Uniform(lower=-1, upper=1), 2), theta=2)
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.4155, -0.173 ,  0.8661]])
        >>> distribution.pdf(samples).round(4)
        array([ 0.2014,  0.3844, 11.8495])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.3764, -0.0596,  0.199 ],
                [-0.1496,  0.115 ,  0.331 ],
                [ 0.0446,  0.2645,  0.444 ]]])

    """

    def __init__(self, dist, theta=2.):
        super(Joe, self).__init__(
            dist=dist,
            trans=joe(len(dist), theta),
            repr_args=[dist, "theta=%s" % theta],
        )



"""Skewness operator."""
import numpoly

from .expected import E
from .standard_deviation import Std


def Skew(poly, dist=None, **kws):
    """
    Skewness operator.

    Element by element 3rd order statistics of a distribution or polynomial.

    Args:
        poly (numpoly.ndpoly, Distribution):
            Input to take skewness on.
        dist (Distribution):
            Defines the space the skewness is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``skewness.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.Skew(dist)
        array([2., 0.])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.Skew(poly, dist)
        array([nan,  2.,  0.,  0.])
    """
    if dist is None:
        dist, poly = poly, numpoly.variable(len(poly))
    poly = numpoly.set_dimensions(poly, len(dist))
    if not poly.isconstant:
        return poly.tonumpy()**3

    poly = poly-E(poly, dist, **kws)
    poly = numpoly.true_divide(poly, Std(poly, dist, **kws))
    return E(poly**3, dist, **kws)



"""Main Sobol sensitivity index."""
import numpy
import numpoly

from ..variance import Var
from ..conditional import E_cond


def Sens_m(poly, dist, **kws):
    """
    Variance-based decomposition/Sobol' indices.

    First order sensitivity indices.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to find first order Sobol indices on.
        dist (Distribution):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray):
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist),) + poly.shape``.

    Examples:
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> chaospy.Sens_m(poly, distribution)
        array([[0.        , 1.        , 0.        , 0.42857143],
               [0.        , 0.        , 1.        , 0.42857143]])
    """
    dim = len(dist)
    poly = numpoly.set_dimensions(poly, dim)

    out = numpy.zeros((dim,) + poly.shape)
    variance = Var(poly, dist, **kws)
    valids = variance != 0

    for idx, unit_vec in enumerate(numpy.eye(dim, dtype=int)):

        conditional = E_cond(poly[valids], unit_vec, dist, **kws)
        out[idx, valids] = Var(conditional, dist, **kws)
        out[idx, valids] /= variance[valids]

    return out



"""
In some cases a constructed distribution that are first and foremost data
driven. In such scenarios it make sense to make use of
`kernel density estimation`_ (KDE). In ``chaospy`` KDE can be accessed through
the :func:`SampleDist` constructor.

Basic usage of the :func:`SampleDist` constructor involves just passing the
data as input argument::

    >>> data = [3, 4, 5, 5]
    >>> distribution = chaospy.SampleDist(data)

This distribution can be used as any other distributions::

    >>> distribution.cdf([3, 3.5, 4, 4.5, 5]).round(4)
    array([0.    , 0.1932, 0.4279, 0.7043, 1.    ])
    >>> distribution.mom(1).round(4)
    4.25
    >>> distribution.sample(4).round(4)
    array([4.4131, 3.3111, 4.9139, 4.1042])

It also supports lower and upper bounds defining where the range is expected to
appear, which gives a slightly different distribution::

    >>> distribution = chaospy.SampleDist(data, lo=2, up=6)
    >>> distribution.cdf([3, 3.5, 4, 4.5, 5]).round(4)
    array([0.1344, 0.2543, 0.4001, 0.5716, 0.7552])

In addition multivariate distributions supported::

    >>> data = [[1, 2, 2, 3], [5, 5, 4, 3]]
    >>> distribution = chaospy.SampleDist(data)
    >>> distribution.sample(4).round(4)
    array([[1.5286, 2.0468, 2.1125, 1.8947],
           [4.402 , 4.1522, 4.4384, 4.5737]])

.. _kernel density estimation: \
https://en.wikipedia.org/wiki/Kernel_density_estimation
"""
import numpy
from scipy.stats import gaussian_kde
import chaospy

from chaospy.distributions import SimpleDistribution


class sample_dist(SimpleDistribution):
    """A distribution that is based on a kernel density estimator (KDE)."""

    def __init__(self, samples, lo, up):
        samples = numpy.asarray(samples)
        self.samples = samples
        self.kernel = gaussian_kde(samples, bw_method="scott")
        self.flo = self.kernel.integrate_box_1d(0, lo)
        self.fup = self.kernel.integrate_box_1d(0, up)
        self.unbound = numpy.all(lo == samples.min())
        self.unbound &= numpy.all(up == samples.max())
        super(sample_dist, self).__init__(
            parameters=dict(lo=lo, up=up),
            repr_args=[repr(samples), lo, up],
        )

    def _cdf(self, xloc, lo, up):
        cdf_vals = numpy.array([self.kernel.integrate_box_1d(0, x)
                                for x in xloc])
        return (cdf_vals-self.flo)/(self.fup-self.flo)

    def _pdf(self, x, lo, up):
        return self.kernel(x)

    def _lower(self, lo, up):
        return lo

    def _upper(self, lo, up):
        return up

    def _mom(self, k, lo, up):
        if self.unbound:
            return numpy.prod(numpy.mean(self.samples.T**k, -1))
        raise chaospy.StochasticallyDependentError("component lack support")


def SampleDist(samples, lo=None, up=None, threshold=1e-5):
    """
    Distribution based on samples.

    Estimates a distribution from the given samples by constructing a kernel
    density estimator (KDE).

    Args:
        samples (numpy.ndarray):
            Sample values to construction of the KDE. Either shape
            ``(N,)`` or ``(D, N)``, where ``N`` are the number of
            samples, and ``D`` is the number of dimension in the
            distribution.
        lo (float):
            Location of lower bound.
        up (float):
            Location of upper bound.
        threshold (float):
            Threshold for how low the correlation between two
            columns should be before defining them as
            stochastically independent.

    Example:
        >>> distribution = chaospy.SampleDist([0, 1, 1, 1, 2])
        >>> distribution
        sample_dist(array([0, 1, 1, 1, 2]), 0, 2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([0.    , 0.6016, 1.    , 1.3984, 2.    ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.2254, 0.4272, 0.5135, 0.4272, 0.2254])
        >>> distribution.sample(4).round(4)
        array([0.3662, 0.6073, 0.9156, 1.0883])
        >>> distribution.mom(1).round(4)
        1.0

    """
    samples = numpy.atleast_2d(samples)
    assert samples.ndim == 2, "samples have too many dimensions provided"

    if lo is None:
        lo = samples.min(axis=-1)
    else:
        lo = numpy.broadcast_to(lo, len(samples))
    if up is None:
        up = samples.max(axis=-1)
    else:
        up = numpy.broadcast_to(up, len(samples))

    # construct vector of marginals
    distributions = []
    for samples_, lo_, up_ in zip(samples, lo, up):
        #construct the kernel density estimator
        try:
            dist = sample_dist(samples_, lo_, up_)
        #raised by gaussian_kde if dataset is singular matrix
        except numpy.linalg.LinAlgError:
            dist = chaospy.Uniform(lower=-numpy.inf, upper=numpy.inf)
        distributions.append(dist)

    if len(samples) == 1:
        distributions = distributions[0]

    else:
        distributions = chaospy.J(*distributions)

        # Attach dependencies to data.
        correlation = numpy.corrcoef(samples)
        correlation[numpy.abs(correlation) <= threshold] = 0
        if numpy.any(correlation != numpy.diag(numpy.diag(correlation))):
            distributions = chaospy.Nataf(distributions, correlation)

    return distributions



"""
Addition operator.

Example usage
-------------

Distribution and a constant::

    >>> distribution = chaospy.Normal(0, 1)+10
    >>> distribution
    Add(Normal(mu=0, sigma=1), 10)
    >>> distribution.sample(5).round(4)
    array([10.395 ,  8.7997, 11.6476,  9.9553, 11.1382])
    >>> distribution.fwd([9, 10, 11]).round(4)
    array([0.1587, 0.5   , 0.8413])
    >>> distribution.inv(distribution.fwd([9, 10, 11])).round(4)
    array([ 9., 10., 11.])
    >>> distribution.pdf([9, 10, 11]).round(4)
    array([0.242 , 0.3989, 0.242 ])
    >>> distribution.mom([1, 2, 3]).round(4)
    array([  10.,  101., 1030.])
    >>> distribution.ttr([1, 2, 3]).round(4)
    array([[10., 10., 10.],
           [ 1.,  2.,  3.]])

Construct joint addition distribution::

    >>> lhs = chaospy.Uniform(2, 3)
    >>> rhs = chaospy.Uniform(3, 4)
    >>> addition = lhs + rhs
    >>> addition
    Add(Uniform(lower=2, upper=3), Uniform(lower=3, upper=4))
    >>> joint1 = chaospy.J(lhs, addition)
    >>> joint2 = chaospy.J(rhs, addition)

Generate random samples::

    >>> joint1.sample(4).round(4)
    array([[2.2123, 2.0407, 2.3972, 2.2331],
           [6.0541, 5.2478, 6.1397, 5.6253]])
    >>> joint2.sample(4).round(4)
    array([[3.1823, 3.7435, 3.0696, 3.8853],
           [6.1349, 6.6747, 5.485 , 5.9143]])

Forward transformations::

    >>> lcorr = numpy.array([2.1, 2.5, 2.9])
    >>> rcorr = numpy.array([3.01, 3.5, 3.99])
    >>> joint1.fwd([lcorr, lcorr+rcorr]).round(4)
    array([[0.1 , 0.5 , 0.9 ],
           [0.01, 0.5 , 0.99]])
    >>> joint2.fwd([rcorr, lcorr+rcorr]).round(4)
    array([[0.01, 0.5 , 0.99],
           [0.1 , 0.5 , 0.9 ]])

Inverse transformations::

    >>> joint1.inv(joint1.fwd([lcorr, lcorr+rcorr])).round(4)
    array([[2.1 , 2.5 , 2.9 ],
           [5.11, 6.  , 6.89]])
    >>> joint2.inv(joint2.fwd([rcorr, lcorr+rcorr])).round(4)
    array([[3.01, 3.5 , 3.99],
           [5.11, 6.  , 6.89]])

Raw moments::

    >>> joint1.mom([(0, 1, 1), (1, 0, 1)]).round(4)
    array([ 6.    ,  2.5   , 15.0833])
    >>> joint2.mom([(0, 1, 1), (1, 0, 1)]).round(4)
    array([ 6.    ,  3.5   , 21.0833])
"""
from __future__ import division
from scipy.special import comb
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Add(OperatorDistribution):
    """Addition operator."""

    def __init__(self, left, right):
        super(Add, self).__init__(
            left=left,
            right=right,
            repr_args=[left, right],
        )

    def _lower(self, idx, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> chaospy.Uniform().lower
            array([0.])
            >>> chaospy.Add(chaospy.Uniform(), 2).lower
            array([2.])
            >>> chaospy.Add(2, chaospy.Uniform()).lower
            array([2.])
        """
        if isinstance(left, Distribution):
            left = left._get_lower(idx, cache=cache)
        if isinstance(right, Distribution):
            right = right._get_lower(idx, cache=cache)
        return left+right

    def _upper(self, idx, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> chaospy.Uniform().upper
            array([1.])
            >>> chaospy.Add(chaospy.Uniform(), 2).upper
            array([3.])
            >>> chaospy.Add(2, chaospy.Uniform()).upper
            array([3.])

        """
        if isinstance(left, Distribution):
            left = left._get_upper(idx, cache=cache)
        if isinstance(right, Distribution):
            right = right._get_upper(idx, cache=cache)
        return (left.T+right.T).T

    def _cdf(self, xloc, idx, left, right, cache):
        if isinstance(right, Distribution):
            left, right = right, left
        xloc = (xloc.T-numpy.asfarray(right).T).T
        uloc = left._get_fwd(xloc, idx, cache=cache)
        return uloc

    def _pdf(self, xloc, idx, left, right, cache):
        """
        Probability density function.

        Example:
            >>> chaospy.Uniform().pdf([-2, 0, 2, 4])
            array([0., 1., 0., 0.])
            >>> chaospy.Add(chaospy.Uniform(), 2).pdf([-2, 0, 2, 4])
            array([0., 0., 1., 0.])
            >>> chaospy.Add(2, chaospy.Uniform()).pdf([-2, 0, 2, 4])
            array([0., 0., 1., 0.])

        """
        if isinstance(right, Distribution):
            left, right = right, left
        xloc = (xloc.T-numpy.asfarray(right).T).T
        return left._get_pdf(xloc, idx, cache=cache)

    def _ppf(self, uloc, idx, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> chaospy.Uniform().inv([0.1, 0.2, 0.9])
            array([0.1, 0.2, 0.9])
            >>> chaospy.Add(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9])
            array([2.1, 2.2, 2.9])
            >>> chaospy.Add(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9])
            array([2.1, 2.2, 2.9])

        """
        if isinstance(right, Distribution):
            left, right = right, left
        xloc = left._get_inv(uloc, idx, cache=cache)
        return (xloc.T+numpy.asfarray(right).T).T

    def _mom(self, keys, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> chaospy.Uniform().mom([0, 1, 2, 3]).round(4)
            array([1.    , 0.5   , 0.3333, 0.25  ])
            >>> chaospy.Add(chaospy.Uniform(), 2).mom([0, 1, 2, 3]).round(4)
            array([ 1.    ,  2.5   ,  6.3333, 16.25  ])
            >>> chaospy.Add(2, chaospy.Uniform()).mom([0, 1, 2, 3]).round(4)
            array([ 1.    ,  2.5   ,  6.3333, 16.25  ])

        """
        del cache
        keys_ = numpy.mgrid[tuple(slice(0, key+1, 1) for key in keys)]
        keys_ = keys_.reshape(len(self), -1)

        if isinstance(left, Distribution):
            if chaospy.shares_dependencies(left, right):
                raise chaospy.StochasticallyDependentError(
                    "%s: left and right side of sum stochastically dependent." % self)
            left = [left._get_mom(key) for key in keys_.T]
        else:
            left = list(reversed(numpy.array(left).T**keys_.T))

        if isinstance(right, Distribution):
            right = [right._get_mom(key) for key in keys_.T]
        else:
            right = list(reversed(numpy.prod(numpy.array(right).T**keys_.T, -1)))

        out = 0.
        for idx in range(keys_.shape[1]):
            key = keys_.T[idx]
            coef = numpy.prod(comb(keys, key))
            out += coef*left[idx]*right[idx]*numpy.all(key <= keys)
        return out

    def _ttr(self, kloc, idx, left, right, cache):
        """
        Three terms recurrence coefficients.

        Example:
            >>> chaospy.Uniform().ttr([0, 1, 2, 3]).round(4)
            array([[ 0.5   ,  0.5   ,  0.5   ,  0.5   ],
                   [-0.    ,  0.0833,  0.0667,  0.0643]])
            >>> chaospy.Add(chaospy.Uniform(), 2).ttr([0, 1, 2, 3]).round(4)
            array([[ 2.5   ,  2.5   ,  2.5   ,  2.5   ],
                   [-0.    ,  0.0833,  0.0667,  0.0643]])
            >>> chaospy.Add(2, chaospy.Uniform()).ttr([0, 1, 2, 3]).round(4)
            array([[ 2.5   ,  2.5   ,  2.5   ,  2.5   ],
                   [-0.    ,  0.0833,  0.0667,  0.0643]])

        """
        del cache
        if isinstance(right, Distribution):
            left, right = right, left
        coeff0, coeff1 = left._get_ttr(kloc, idx)
        return coeff0+numpy.asarray(right), coeff1


def add(left, right):
    return Add(left, right)



"""Standard deviation."""
import numpy

from .variance import Var


def Std(poly, dist=None, **kws):
    """
    Element by element 2nd order statistics.

    Args:
        poly (numpoly.ndpoly, Distribution):
            Input to take variance on.
        dist (Distribution):
            Defines the space the variance is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``variation.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.Std(dist)
        array([1., 2.])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.Std(poly, dist)
        array([ 0.        ,  1.        ,  2.        , 28.28427125])

    """
    return numpy.sqrt(Var(poly, dist=dist, **kws))



"""
Negative of a distribution.

Example usage
-------------

Invert sign of a distribution::

    >>> distribution = -chaospy.Uniform()
    >>> distribution
    Negative(Uniform())
    >>> distribution.sample(5).round(4)
    array([-0.3464, -0.885 , -0.0497, -0.5178, -0.1275])
    >>> distribution.fwd([-0.3, -0.2, -0.1])
    array([0.7, 0.8, 0.9])
    >>> distribution.inv(distribution.fwd([-0.3, -0.2, -0.1]))
    array([-0.3, -0.2, -0.1])
    >>> distribution.pdf([-0.3, -0.2, -0.1])
    array([1., 1., 1.])
    >>> distribution.mom([1, 2, 3]).round(4)
    array([-0.5   ,  0.3333, -0.25  ])
    >>> distribution.ttr([1, 2, 3]).round(4)
    array([[-0.5   , -0.5   , -0.5   ],
           [ 0.0833,  0.0667,  0.0643]])

"""
import numpy

from ..baseclass import Distribution, OperatorDistribution


class Negative(OperatorDistribution):
    """Negative of a distribution."""

    def __init__(self, dist):
        """
        Constructor.

        Args:
            dist (Distribution) : distribution.
        """
        super(Negative, self).__init__(
            left=dist,
            right=0,
        )
        self._repr_args = [dist]

    def _lower(self, idx, left, right, cache):
        del right
        return -left._get_upper(idx, cache)

    def _upper(self, idx, left, right, cache):
        del right
        return -left._get_lower(idx, cache)

    def _pdf(self, xloc, idx, left, right, cache):
        del right
        return left._get_pdf(-xloc, idx, cache)

    def _cdf(self, xloc, idx, left, right, cache):
        del right
        return 1-left._get_fwd(-xloc, idx, cache)

    def _ppf(self, uloc, idx, left, right, cache):
        del right
        return -left._get_inv(1-uloc, idx, cache)

    def _mom(self, kloc, left, right, cache):
        """Statistical moments."""
        del right
        del cache
        return (-1)**numpy.sum(kloc)*left._get_mom(kloc)

    def _ttr(self, kloc, idx, left, right, cache):
        """Three terms recurrence coefficients."""
        del right
        del cache
        alpha, beta = left._get_ttr(kloc, idx)
        return -alpha, beta


def neg(left):
    """
    Negative of a distribution.

    Args:
        dist (Distribution) : distribution.
    """
    return Negative(left)



"""Triangle probability distribution."""
import numpy
from scipy import special, misc

from ..baseclass import SimpleDistribution, LowerUpperDistribution
from .beta import beta_


class triangle(SimpleDistribution):
    """Triangle probability distribution."""

    def __init__(self, a=.5):
        # assert numpy.all(a>=0) and numpy.all(a<=1)
        super(triangle, self).__init__(dict(a=a))

    def _pdf(self, xloc, a):
        return numpy.where(xloc < a, 2*xloc/a, 2*(1-xloc)/(1-a))

    def _cdf(self, xloc, a):
        return numpy.where(xloc < a, xloc**2/(a+(a == 0)),
                (2*xloc-xloc*xloc-a)/(1-a+(a == 1)))

    def _ppf(self, qloc, a):
        return numpy.where(qloc<a, numpy.sqrt(qloc*a), 1-numpy.sqrt(1-a-qloc*(1-a)))

    def _mom(self, k, a):
        a_ = a*(a!=1)
        out = 2*(1.-a_**(k+1))/((k+1)*(k+2)*(1-a_))
        return numpy.where(a==1, 2./(k+2), out)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 1.

    def _ttr(self, k, a):
        if a == 0:
            return beta_()._ttr(k, 1, 2)
        if a == 1:
            return beta_()._ttr(k, 2, 1)

        from ...quadrature import quad_fejer, discretized_stieltjes
        q1, w1 = quad_fejer(int(1000*a), (0, a))
        q2, w2 = quad_fejer(int(1000*(1-a)), (a, 1))
        qloc = numpy.concatenate([q1,q2], 1)
        w = numpy.concatenate([w1,w2])*self._pdf(qloc[0], a)

        coeffs, _, _ = discretized_stieltjes(k, qloc, w)
        return coeffs[:, 0, -1]


class Triangle(LowerUpperDistribution):
    """
    Triangle Distribution.

    Must have lower <= midpoint <= upper.

    Args:
        lower (float, Distribution):
            Lower bound
        midpoint (float, Distribution):
            Location of the top
        upper (float, Distribution):
            Upper bound

    Examples:
        >>> distribution = chaospy.Triangle(-1, 0 ,1)
        >>> distribution
        Triangle(-1, 0.5, 1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-1.   , -0.368, -0.106,  0.106,  0.368,  1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.632, 0.894, 0.894, 0.632, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.168, -0.52 ,  0.685, -0.018])
        >>> distribution.mom(1).round(4)
        0.0
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[-0.    ,  0.    , -0.    ,  0.    ],
               [ 4.    ,  0.1667,  0.2333,  0.2327]])

    """

    def __init__(self, lower=-1, midpoint=0, upper=1):
        midpoint = (midpoint-lower)*1./(upper-lower)
        super(Triangle, self).__init__(
            dist=triangle(midpoint),
            lower=lower,
            upper=upper,
            repr_args=["midpoint=%s" % midpoint],
        )
        self._repr_args = [lower, midpoint, upper]



"""Weibull Distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class weibull(SimpleDistribution):
    """Weibull Distribution."""

    def __init__(self, a=1):
        super(weibull, self).__init__(dict(a=a))

    def _pdf(self, x, a):
        return a*x**(a-1)*numpy.e**(-x**a)

    def _cdf(self, x, a):
        return (1-numpy.e**(-x**a))

    def _ppf(self, q, a):
        return (-numpy.log(1-q+1*(q==1)))**(1./a)*(q!=1)+30.**(1./a)*(q==1)

    def _mom(self, k, a):
        return special.gamma(1.+k*1./a)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 30.**(1./a)


class Weibull(ShiftScaleDistribution):
    """
    Weibull Distribution

    Args:
        shape (float, Distribution):
            Shape parameter.
        scale (float, Distribution):
            Scale parameter.
        shift (float, Distribution):
            Location of lower bound.

    Examples:
        >>> distribution = chaospy.Weibull(2)
        >>> distribution
        Weibull(2)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.472, 0.715, 0.957, 1.269, 5.477])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.756, 0.858, 0.766, 0.507, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.03 , 0.35 , 1.732, 0.811])
        >>> distribution.mom(1).round(4)
        0.8862

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Weibull, self).__init__(
            dist=weibull(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Truncated normal distribution."""
import numpy
from scipy import special
import chaospy

from .normal import normal
from ..baseclass import SimpleDistribution
from ..operators import J



class trunc_normal(SimpleDistribution):

    def __init__(self, lower=-1, upper=1, mu=0, sigma=1):
        super(trunc_normal, self).__init__(
            parameters=dict(a=lower, b=upper, mu=mu, sigma=sigma),
            repr_args=["lower=%s" % lower, "upper=%s" % upper,
                       "mu=%s" % mu, "sigma=%s" % sigma],
        )

    def _pdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = (x-mu)/sigma
        norm = (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)
        return norm/(fb-fa)

    def _cdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = special.ndtr((x-mu)/sigma)
        return (x-fa)/(fb-fa)

    def _ppf(self, q, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        return special.ndtri(q*(fb-fa) + fa)*sigma + mu

    def _lower(self, a, b, mu, sigma):
        del b
        lower = normal()._lower()*sigma+mu
        return numpy.where(a < lower, lower, a)

    def _upper(self, a, b, mu, sigma):
        del a
        upper = normal()._upper()*sigma+mu
        return numpy.where(b > upper, upper, b)


class TruncNormal(J):
    """
    Truncated normal distribution

    Args:
        lower (float, Distribution):
            Location of lower threshold
        upper (float, Distribution):
            Location of upper threshold
        mu (float, Distribution):
            Mean of normal distribution
        sigma (float, Distribution):
            Standard deviation of normal distribution

    Examples:
        >>> full_trunc = chaospy.TruncNormal(lower=-1, upper=1)
        >>> half_trunc = chaospy.TruncNormal(upper=1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> half_trunc.inv(uloc).round(3)
        array([-8.22 , -0.961, -0.422,  0.012,  0.448,  1.   ])
        >>> xloc = full_trunc.inv(uloc)
        >>> xloc.round(3)
        array([-1.   , -0.538, -0.172,  0.172,  0.538,  1.   ])
        >>> numpy.allclose(full_trunc.fwd(xloc), uloc)
        True
        >>> full_trunc.pdf(xloc).round(3)
        array([0.354, 0.506, 0.576, 0.576, 0.506, 0.354])
        >>> half_trunc.pdf(xloc).round(3)
        array([0.288, 0.41 , 0.467, 0.467, 0.41 , 0.288])
        >>> full_trunc.sample(4).round(3)
        array([ 0.266, -0.715,  0.868, -0.03 ])
        >>> half_trunc.sample(4).round(3)
        array([ 0.625, -0.921, -1.822, -0.428])

    """

    def __init__(self, lower=-numpy.inf, upper=numpy.inf, mu=0, sigma=1):
        super(TruncNormal, self).__init__(
            trunc_normal(lower=lower, upper=upper, mu=mu, sigma=sigma))
        self._repr_args = chaospy.format_repr_kwargs(lower=(lower, -numpy.inf))
        self._repr_args += chaospy.format_repr_kwargs(upper=(upper, numpy.inf))
        self._repr_args += chaospy.format_repr_kwargs(lower=(mu, 0))
        self._repr_args += chaospy.format_repr_kwargs(lower=(sigma, 1))



"""Non-central Chi-squared distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class chi_squared(SimpleDistribution):
    """Central Chi-squared distribution."""

    def __init__(self, df, nc):
        super(chi_squared, self).__init__(dict(df=df, nc=nc))

    def _pdf(self, x, df, nc):
        output = 0.5*numpy.e**(-0.5*(x+nc))
        output *= (x/nc)**(0.25*df-0.5)
        output *= special.iv(0.5*df-1, (nc*x)**0.5)
        return output

    def _cdf(self, x, df, nc):
        return special.chndtr(x, df, nc)

    def _ppf(self, qloc, df, nc):
        qloc = numpy.clip(qloc, None, 1-1e-12)
        return special.chndtrix(qloc, df, nc)

    def _lower(self, df, nc):
        return 0.

    def _upper(self, df, nc):
        return special.chndtrix(1-1e-12, df, nc)


class ChiSquared(ShiftScaleDistribution):
    """
    (Non-central) Chi-squared distribution.

    Args:
        df (float, Distribution):
            Degrees of freedom
        nc (float, Distribution):
            Non-centrality parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.ChiSquared(df=15)
        >>> distribution
        ChiSquared(15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   , 11.003, 13.905, 16.784, 20.592, 95.358])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.062, 0.072, 0.064, 0.041, 0.   ])
        >>> distribution.sample(4).round(3)
        array([17.655,  9.454, 26.66 , 15.047])

    """

    def __init__(self, df=1, nc=1, scale=1, shift=0):
        super(ChiSquared, self).__init__(
            dist=chi_squared(df, nc),
            scale=scale,
            shift=shift,
            repr_args=[df]+chaospy.format_repr_kwargs(nc=(nc, 1)),
        )



import numpy
from scipy.special import comb
import numpoly
import chaospy

from .distribution import Distribution


class LowerUpperDistribution(Distribution):
    """
    Lower-Upper transformation.

    Linear transforms any distribution on any interval `[lower, upper]`.

    Args:
        dist (Distribution):
            The underlying distribution to be scaled.
        lower (float, Sequence[float], Distribution):
            Lower bounds.
        upper (float, Sequence[float], Distribution):
            Upper bounds.

    Attributes:
        shift (numpy.ndarray):
            The shift of the distribution.
        scale (numpy.ndarray):
            The scale of the distribution.

    """

    def __init__(
            self,
            dist,
            lower=0.,
            upper=1.,
            rotation=None,
            repr_args=None,
    ):
        assert isinstance(dist, Distribution), "'dist' should be a distribution"
        if repr_args is None:
            repr_args = dist._repr_args[:]
        repr_args += chaospy.format_repr_kwargs(lower=(lower, 0), upper=(upper, 1))

        dependencies, parameters, rotation, = chaospy.declare_dependencies(
            distribution=self,
            parameters=dict(lower=lower, upper=upper),
            rotation=rotation,
        )
        super(LowerUpperDistribution, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            repr_args=repr_args,
        )
        self._dist = dist

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(LowerUpperDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        lower = parameters["lower"]
        if isinstance(lower, Distribution):
            lower = lower._get_cache(idx, cache, get=0)
        if idx is not None and len(lower) > 1:
            lower = lower[idx]
        upper = parameters["upper"]
        if isinstance(upper, Distribution):
            upper = upper._get_cache(idx, cache, get=0)
        if idx is not None and len(upper) > 1:
            upper = upper[idx]
        assert not assert_numerical or not (isinstance(lower, Distribution) or
                                            isinstance(upper, Distribution))

        lower0 = self._dist._get_lower(idx, cache.copy())
        upper0 = self._dist._get_upper(idx, cache.copy())
        scale = (upper-lower)/(upper0-lower0)
        shift = lower-lower0*(upper-lower)/(upper0-lower0)
        parameters = self._dist.get_parameters(idx, cache, assert_numerical=assert_numerical)
        return dict(dist=self._dist, scale=scale, shift=shift, parameters=parameters)

    def _lower(self, dist, scale, shift, parameters):
        return dist._lower(**parameters)*scale+shift

    def _upper(self, dist, scale, shift, parameters):
        return dist._upper(**parameters)*scale+shift

    def _ppf(self, qloc, dist, scale, shift, parameters):
        return dist._ppf(qloc, **parameters)*scale+shift

    def _cdf(self, xloc, dist, scale, shift, parameters):
        return dist._cdf((xloc-shift)/scale, **parameters)

    def _pdf(self, xloc, dist, scale, shift, parameters):
        return dist._pdf((xloc-shift)/scale, **parameters)/scale

    def _mom(self, kloc, dist, scale, shift, parameters):
        poly = numpoly.variable(len(self))
        poly = numpoly.sum(scale*poly, axis=-1)+shift
        poly = numpoly.set_dimensions(numpoly.prod(poly**kloc), len(self))
        out = sum(dist._get_mom(key)*coeff
                  for key, coeff in zip(poly.exponents, poly.coefficients))
        return out

    def _ttr(self, kloc, dist, scale, shift, parameters):
        coeff0, coeff1 = dist._ttr(kloc, **parameters)
        coeff0 = coeff0*scale+shift
        coeff1 = coeff1*scale*scale
        return coeff0, coeff1



"""Double Weibull distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class double_weibull(SimpleDistribution):
    """Double weibull distribution."""

    def __init__(self, c):
        super(double_weibull, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        ax = numpy.abs(x)
        Px = c/2.0*ax**(c-1.0)*numpy.exp(-ax**c)
        return Px

    def _cdf(self, x, c):
        Cx1 = 0.5*numpy.exp(-abs(x)**c)
        return numpy.where(x > 0, 1-Cx1, Cx1)

    def _ppf(self, q, c):
        q_ = numpy.where(q > .5, 1-q, q)
        Cq1 = numpy.where(q_ == 0, self._upper(c), 1)
        Cq1 = numpy.where((q != 0) & (c != 0),
                          (-numpy.log(2*q_))**(1./c), Cq1)
        return numpy.where(q > .5, Cq1, -Cq1)

    def _lower(self, c):
        return -(-numpy.log(2e-10))**(1./c)

    def _upper(self, c):
        return (-numpy.log(2e-10))**(1./c)


class DoubleWeibull(ShiftScaleDistribution):
    """
    Double Weibull distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.DoubleWeibull(1.5)
        >>> distribution
        DoubleWeibull(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-7.93 , -0.943, -0.368,  0.368,  0.943,  7.93 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.291, 0.364, 0.364, 0.291, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.513, -1.293,  1.747, -0.11 ])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(DoubleWeibull, self).__init__(
            dist=double_weibull(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Discrete uniform probability distribution."""
import numpy

from ..baseclass import SimpleDistribution
from ..operators import J


class DiscreteUniform(SimpleDistribution):
    """
    Discrete uniform probability distribution.

    Args:
        lower (float, chaospy.Distribution):
            Lower threshold of distribution. Must be smaller than ``upper``.
            Value will be rounded up to closes integer.
        upper (float, chaospy.Distribution):
            Upper threshold of distribution. Value will be rouned down to
            closes integer.

    Examples:
        >>> distribution = chaospy.DiscreteUniform(2, 4)
        >>> distribution
        DiscreteUniform(2, 4)
        >>> qloc = numpy.linspace(0, 1, 9)
        >>> qloc.round(2)
        array([0.  , 0.12, 0.25, 0.38, 0.5 , 0.62, 0.75, 0.88, 1.  ])
        >>> distribution.inv(qloc).round(2)
        array([1.5 , 1.88, 2.25, 2.62, 3.  , 3.38, 3.75, 4.12, 4.5 ])
        >>> distribution.fwd(distribution.inv(qloc)).round(2)
        array([0.  , 0.12, 0.25, 0.38, 0.5 , 0.62, 0.75, 0.88, 1.  ])
        >>> distribution.pdf(distribution.inv(qloc)).round(4)
        array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5])
        >>> distribution.sample(4)
        array([3, 2, 4, 3])
        >>> distribution.mom(1).round(4)
        3.0
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[3.    , 3.    , 3.    , 3.3333],
               [1.    , 0.6667, 0.3333, 0.    ]])

    """
    interpret_as_integer = True

    def __init__(self, lower, upper):
        super(DiscreteUniform, self).__init__(dict(lower=lower, upper=upper))
        self._repr_args = [lower, upper]

    def _cdf(self, x_data, lower, upper):
        """Cumulative distribution function."""
        lower = numpy.round(lower)
        upper = numpy.round(upper)
        out = (x_data-lower+0.5)/(upper-lower+1)
        return out

    def _lower(self, lower, upper):
        """Lower bounds."""
        return numpy.round(lower)-0.5

    def _upper(self, lower, upper):
        """Upper bounds."""
        return numpy.round(upper)+0.5

    def _pdf(self, x_data, lower, upper):
        """Probability density function."""
        return x_data**0/(numpy.round(upper)-numpy.round(lower))

    def _ppf(self, q_data, lower, upper):
        """Point percentile function."""
        lower = numpy.round(lower)
        upper = numpy.rint(upper)
        return q_data*(upper-lower+1)+lower-0.5

    def _mom(self, k_data, lower, upper):
        """Raw statistical moments."""
        return numpy.mean(numpy.arange(
            numpy.ceil(lower), numpy.floor(upper)+1)**k_data)

    def _ttr(self, k_data, lower, upper):
        """Three terms recurrence coefficients."""
        from chaospy.quadrature import discretized_stieltjes
        abscissas = numpy.arange(numpy.ceil(lower), numpy.floor(upper)+1)
        weights = numpy.repeat(1./len(abscissas), len(abscissas))
        (alpha, beta), _, _ = discretized_stieltjes(k_data, [abscissas], weights)
        return alpha[0, -1], beta[0, -1]



"""Inverse Gamma distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class inverse_gamma(SimpleDistribution):

    def __init__(self, a):
        super(inverse_gamma, self).__init__(dict(a=a))

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 1./special.gammainccinv(a, 1-1e-10)

    def _pdf(self, x, a):
        x_ = numpy.where(x, x, 1)
        return numpy.where(x, x_**(-a-1)*numpy.exp(-1./x_)/special.gamma(a), 0)

    def _cdf(self, x, a):
        return numpy.where(x, special.gammaincc(a, 1./numpy.where(x, x, 1)), 0)

    def _ppf(self, q, a):
        return 1./special.gammainccinv(a, q)

    def _mom(self, k, a):
        if k > a:
            return self._upper(a)
        return numpy.prod(a-numpy.arange(1, k.item()+1))


class InverseGamma(ShiftScaleDistribution):
    """
    Inverse-Gamma distribution.

    Args:
        shape (float, Distribution):
            Shape parameter. a>0.
        scale (float, Distribution):
            Scale parameter. scale!=0
        shift (float, Distribution):
            Location of the lower bound.

    Examples:
        >>> distribution = chaospy.InverseGamma(shape=5)
        >>> distribution
        InverseGamma(5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  0.149,  0.191,  0.241,  0.324, 38.218])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 4.629, 4.569, 3.352, 1.65 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.258, 0.129, 0.508, 0.21 ])
        >>> distribution.mom(1).round(3)
        4.0

    """

    def __init__(self, shape, scale=1, shift=0):
        super(InverseGamma, self).__init__(
            dist=inverse_gamma(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Laplace Probability Distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class laplace(SimpleDistribution):
    """Laplace Probability Distribution."""

    def __init__(self):
        super(laplace, self).__init__()

    def _pdf(self, x):
        return numpy.e**-numpy.abs(x)/2

    def _cdf(self, x):
        return (1+numpy.sign(x)*(1-numpy.e**-abs(x)))/2

    def _mom(self, k):
        return special.factorial(k)*((k+1)%2)

    def _ppf(self, x):
        return numpy.where(x > .5, -numpy.log(2*(1-x)), numpy.log(2*x))

    def _lower(self):
        return -20.

    def _upper(self):
        return 20.

    def _ttr(self, k):
        q1, w1 = chaospy.quad_fejer(500, (-20, 0))
        q2, w2 = chaospy.quad_fejer(500, (0, 20))
        qloc = numpy.concatenate([q1,q2], 1)
        w = numpy.concatenate([w1,w2])*self._pdf(qloc[0])
        coeffs, _, _ = chaospy.discretized_stieltjes(k, qloc, w)
        return coeffs[:, 0, -1]


class Laplace(ShiftScaleDistribution):
    R"""
    Laplace Probability Distribution

    Args:
        mu (float, Distribution):
            Mean of the distribution.
        scale (float, Distribution):
            Scaling parameter. scale > 0.

    Examples:
        >>> distribution = chaospy.Laplace()
        >>> distribution
        Laplace()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-20.   ,  -0.916,  -0.223,   0.223,   0.916,  20.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0. , 0.2, 0.4, 0.4, 0.2, 0. ])
        >>> distribution.sample(4).round(3)
        array([ 0.367, -1.47 ,  2.308, -0.036])
        >>> distribution.mom(1).round(3)
        0.0
        >>> distribution.ttr([0, 1, 2, 3]).round(3)
        array([[ 0.   , -0.   ,  0.   ,  0.   ],
               [ 1.   ,  2.   , 10.   , 21.592]])

    """
    def __init__(self, mu=0, sigma=1):
        super(Laplace, self).__init__(
            dist=laplace(),
            scale=sigma,
            shift=mu,
        )
        self._repr_args = (chaospy.format_repr_kwargs(mu=(mu, 0))+
                           chaospy.format_repr_kwargs(sigma=(sigma, 1)))



"""Total Sobol sensitivity index."""
import numpy
import numpoly

from ..conditional import E_cond
from ..variance import Var


def Sens_t(poly, dist, **kws):
    """
    Variance-based decomposition
    AKA Sobol' indices

    Total effect sensitivity index

    Args:
        poly (numpoly.ndpoly):
            Polynomial to find first order Sobol indices on.
        dist (Distribution):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray) :
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist),) + poly.shape``.

    Examples:
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> chaospy.Sens_t(poly, dist)
        array([[0.        , 1.        , 0.        , 0.57142857],
               [0.        , 0.        , 1.        , 0.57142857]])
    """
    dim = len(dist)
    poly = numpoly.set_dimensions(poly, dim)

    out = numpy.zeros((dim,)+poly.shape, dtype=float)
    variance = Var(poly, dist, **kws)

    valids = variance != 0
    if not numpy.all(valids):
        out[:, valids] = Sens_t(poly[valids], dist, **kws)
        return out

    out[:] = variance
    for idx, unit_vec in enumerate(numpy.eye(dim, dtype=int)):
        conditional = E_cond(poly, 1-unit_vec, dist, **kws)
        out[idx] -= Var(conditional, dist, **kws)
        out[idx] /= variance

    return out



"""Hyperbolic secant distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class hyperbolic_secant(SimpleDistribution):
    """Hyperbolic secant distribution."""

    def __init__(self):
        super(hyperbolic_secant, self).__init__()

    def _pdf(self, x):
        return .5*numpy.cosh(numpy.pi*x/2.)**-1

    def _cdf(self, x):
        return 2/numpy.pi*numpy.arctan(numpy.e**(numpy.pi*x/2.))

    def _ppf(self, q):
        return 2/numpy.pi*numpy.log(numpy.tan(numpy.pi*q/2.))

    def _lower(self):
        return -21.7

    def _upper(self):
        return 21.7

    def _mom(self, k):
        return numpy.abs(special.euler(k.item())[-1])


class HyperbolicSecant(ShiftScaleDistribution):
    """
    Hyperbolic secant distribution

    Args:
        scale (float, Distribution):
            Scale parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.HyperbolicSecant()
        >>> distribution
        HyperbolicSecant()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-21.7  ,  -0.716,  -0.203,   0.203,   0.716,  21.7  ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.294, 0.476, 0.476, 0.294, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.32 , -1.082,  1.622, -0.036])
        >>> distribution.mom(2).round(3)
        1.0

    """

    def __init__(self, scale=1, shift=0):
        super(HyperbolicSecant, self).__init__(
            dist=hyperbolic_secant(),
            scale=scale,
            shift=shift,
        )



"""Pareto type 1 distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class pareto1(SimpleDistribution):
    """Pareto type 1 distribution."""

    def __init__(self, b):
        super(pareto1, self).__init__(dict(b=b))

    def _pdf(self, x, b):
        return b*x**(-b-1)

    def _cdf(self, x, b):
        return 1-x**-b

    def _ppf(self, q, b):
        return pow(1-q, -1./b)

    def _lower(self, b):
        return 1.

    def _upper(self, b):
        return pow(1e12, 1./b)


class Pareto1(ShiftScaleDistribution):
    """
    Pareto type 1 distribution.

    Lower threshold at scale+loc and survival: x^-shape

    Args:
        shape (float, Distribution):
            Tail index parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Pareto1(15)
        >>> distribution
        Pareto1(15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([1.   , 1.015, 1.035, 1.063, 1.113, 6.31 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([15.   , 11.823,  8.699,  5.644,  2.695,  0.   ])
        >>> distribution.sample(4).round(3)
        array([1.073, 1.008, 1.222, 1.045])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Pareto1, self).__init__(
            dist=pareto1(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""
Frontend for the Hermite Genz-Keister quadrature rule.
"""
import numpy
import scipy.special

from ..combine import combine
from .gk16 import quad_genz_keister_16
from .gk18 import quad_genz_keister_18
from .gk22 import quad_genz_keister_22
from .gk24 import quad_genz_keister_24

GENS_KEISTER_FUNCTIONS = {
    16: quad_genz_keister_16,
    18: quad_genz_keister_18,
    22: quad_genz_keister_22,
    24: quad_genz_keister_24,
}


def quad_genz_keister(order, dist, rule=24):
    """
    Genz-Keister quadrature rule.

    Eabsicassample:
        >>> abscissas, weights = quad_genz_keister(
        ...         order=1, dist=chaospy.Uniform(0, 1))
        >>> abscissas.round(4)
        array([[0.0416, 0.5   , 0.9584]])
        >>> weights.round(4)
        array([0.1667, 0.6667, 0.1667])
    """
    assert isinstance(rule, int)

    if len(dist) > 1:

        if isinstance(order, int):
            values = [quad_genz_keister(order, d, rule) for d in dist]
        else:
            values = [quad_genz_keister(order[i], dist[i], rule)
                      for i in range(len(dist))]

        abscissas = [_[0][0] for _ in values]
        abscissas = combine(abscissas).T
        weights = [_[1] for _ in values]
        weights = numpy.prod(combine(weights), -1)

        return abscissas, weights

    foo = GENS_KEISTER_FUNCTIONS[rule]
    abscissas, weights = foo(order)
    abscissas = dist.inv(scipy.special.ndtr(abscissas))
    abscissas = abscissas.reshape(1, abscissas.size)

    return abscissas, weights



"""Covariance matrix."""
import numpy
import numpoly

from .expected import E
from .. import distributions


def Cov(poly, dist=None, **kws):
    """
    Covariance matrix, or 2rd order statistics.

    Args:
        poly (numpoly.ndpoly, Distribution) :
            Input to take covariance on. Must have `len(poly)>=2`.
        dist (Distribution) :
            Defines the space the covariance is taken on.  It is ignored if
            `poly` is a distribution.

    Returns:
        (numpy.ndarray):
            Covariance matrix with shape ``poly.shape+poly.shape``.

    Examples:
        >>> dist = chaospy.MvNormal([0, 0], [[2, .5], [.5, 1]])
        >>> chaospy.Cov(dist)
        array([[2. , 0.5],
               [0.5, 1. ]])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.Cov(poly, dist)
        array([[  0. ,   0. ,   0. ,   0. ],
               [  0. ,   2. ,   0.5,   0. ],
               [  0. ,   0.5,   1. ,   0. ],
               [  0. ,   0. ,   0. , 225. ]])

    """
    if dist is None:
        dist, poly = poly, numpoly.variable(len(poly))
    poly = numpoly.set_dimensions(poly, len(dist))
    if not poly.isconstant:
        return poly.tonumpy()**2
    poly = poly-E(poly, dist)
    poly = numpoly.outer(poly, poly)
    return E(poly, dist)



"""
Point collocation method, or regression based polynomial chaos expansion builds
open the idea of fitting a polynomial chaos expansion to a set of generated
samples and evaluations. The experiment can be done as follows:

- Select a :ref:`distributions`::

    >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)

- Generate :ref:`orthogonality`::

    >>> orthogonal_expansion = chaospy.orth_ttr(2, distribution)
    >>> orthogonal_expansion
    polynomial([1.0, q1, q0, q1**2-1.0, q0*q1, q0**2-1.0])

- Generate samples using :ref:`sampling` (or alternative abscissas from
  :ref:`quadrature`)::

    >>> samples = distribution.sample(
    ...     2*len(orthogonal_expansion), rule="hammersley")
    >>> samples[:, :4]
    array([[ 0.67448975, -1.15034938,  0.31863936, -0.31863936],
           [-1.42607687, -1.02007623, -0.73631592, -0.50240222]])

- A function evaluated using the nodes generated in the second step::

    >>> def model_solver(param):
    ...     return [param[0]*param[1], param[0]*numpy.e**-param[1]+1]
    >>> solves = numpy.array([model_solver(sample) for sample in samples.T])
    >>> solves[:4].round(8)
    array([[-0.96187423,  3.80745414],
           [ 1.17344406, -2.19038608],
           [-0.23461924,  1.66539168],
           [ 0.16008512,  0.47338898]])

- Bring it all together using `~chaospy.regression.fit_regression`::

    >>> approx_model = chaospy.fit_regression(
    ...      orthogonal_expansion, samples, solves)
    >>> approx_model.round(2)
    polynomial([q0*q1, 0.11*q1**2-1.44*q0*q1+0.05*q0**2-0.09*q1+1.22*q0+0.94])

In this example, the number of collocation points is selected to be twice the
number of unknown coefficients :math:`N+1`. Changing this is obviously
possible. When the number of parameter is equal the number of unknown, the, the
polynomial approximation becomes an interpolation method and overlap with
Lagrange polynomials. If the number of samples are fewer than the number of
unknown, classical least squares can not be used. Instead it possible to use
methods for doing estimation with too few samples.
"""
import logging
import numpy
from scipy import linalg

import numpoly
import chaospy


def fit_regression(
        polynomials,
        abscissas,
        evals,
        model=None,
        retall=False,
):
    """
    Fit a polynomial chaos expansion using linear regression.

    Args:
        polynomials (numpoly.ndpoly):
            Polynomial expansion with ``polynomials.shape == (M,)`` and
            `polynomials.dim=D`.
        abscissas (numpy.ndarray):
            Collocation nodes with ``abscissas.shape == (D, K)``.
        evals (numpy.ndarray):
            Model evaluations with ``len(evals) == K``.
        model (Optional[sklearn.base.BaseEstimator]):
            By default regression is done using the classical least-square
            method. However, if provided, and `sklearn` regression model can be
            used instead.
        retall (bool):
            If True return Fourier coefficients in addition to R.

    Returns:
        (chaospy.ndpoly, numpy.ndarray):
            Fitted polynomial with ``R.shape=evals.shape[1:]`` and ``R.dim=D``.
            The Fourier coefficients in the estimation.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> polynomials = chaospy.polynomial([1, x, y])
        >>> abscissas = [[-1,-1,1,1], [-1,1,-1,1]]
        >>> evals = [0,1,1,2]
        >>> chaospy.fit_regression(polynomials, abscissas, evals).round(14)
        polynomial(0.5*q1+0.5*q0+1.0)

    """
    logger = logging.getLogger(__name__)
    abscissas = numpy.asarray(abscissas)
    if len(abscissas.shape) == 1:
        abscissas = abscissas.reshape(1, *abscissas.shape)
    evals = numpy.array(evals)

    poly_evals = polynomials(*abscissas).T
    shape = evals.shape[1:]
    if shape:
        evals = evals.reshape(evals.shape[0], int(numpy.prod(evals.shape[1:])))

    if model is None:
        uhat = linalg.lstsq(poly_evals, evals)[0]

    else:
        try:
            from sklearn.base import BaseEstimator
        except ImportError:
            raise ValueError(
                "arg model != None requires that scikit-learn is installed")

        if not isinstance(model, BaseEstimator):
            raise ValueError("model not recognized; "
                             "Optional[sklearn.base.BaseEstimator] expected")
        if hasattr(model, "fit_intercept"):
            assert not model.fit_intercept, (
                "model %s must have fit_intercept=False" % model.__class__.__name__)
        uhat = model.fit(poly_evals, evals).coef_.T

    if shape:
        evals = evals.reshape(evals.shape[0], *shape)

    approx_model = numpoly.sum((polynomials*uhat.T), -1)
    approx_model = approx_model.reshape(shape)

    if retall == 1:
        return approx_model, uhat
    if retall == 2:
        return approx_model, uhat, poly_evals
    return approx_model



"""Nakagami-m distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class nakagami(SimpleDistribution):
    """Nakagami-m distribution."""

    def __init__(self, nu):
        super(nakagami, self).__init__(dict(nu=nu))

    def _pdf(self, x, nu):
        return 2*nu**nu/special.gamma(nu)*(x**(2*nu-1.0))*numpy.exp(-nu*x*x)

    def _cdf(self, x, nu):
        return special.gammainc(nu,nu*x*x)

    def _ppf(self, q, nu):
        return numpy.sqrt(1.0/nu*special.gammaincinv(nu, q))

    def _lower(self, nu):
        return 0.

    def _upper(self, nu):
        return numpy.sqrt(1.0/nu*special.gammaincinv(nu, 1-1e-16))


class Nakagami(ShiftScaleDistribution):
    """
    Nakagami-m distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Nakagami(1.5)
        >>> distribution
        Nakagami(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.579, 0.789, 0.991, 1.244, 5.079])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.84 , 1.015, 0.933, 0.63 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.05 , 0.465, 1.615, 0.87 ])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Nakagami, self).__init__(
            dist=nakagami(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Logarithm of another distribution."""
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Logn(OperatorDistribution):
    """
    Logarithm with base N.

    Args:
        dist (Distribution):
            Distribution to perform transformation on.
        base (int, float):
            the logarithm base.

    Example:
        >>> distribution = chaospy.Logn(chaospy.Uniform(1, 2), 3)
        >>> distribution
        Logn(Uniform(lower=1, upper=2), 3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.166 , 0.3063, 0.4278, 0.535 ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.3183, 1.5381, 1.7578, 1.9775])
        >>> distribution.sample(4).round(4)
        array([0.4578, 0.0991, 0.608 , 0.3582])
        >>> distribution.mom(1).round(4)
        0.3516

    """

    def __init__(self, dist, base=2):
        assert isinstance(dist, Distribution)
        assert numpy.all(dist.lower > 0)
        assert base > 0 and base != 1
        super(Logn, self).__init__(
            left=dist,
            right=base,
            repr_args=[dist, base],
        )

    def _lower(self, idx, left, right, cache):
        return numpy.log(left._get_lower(idx, cache))/numpy.log(right)

    def _upper(self, idx, left, right, cache):
        return numpy.log(left._get_upper(idx, cache))/numpy.log(right)

    def _pdf(self, xloc, idx, left, right, cache):
        return left._get_pdf(right**xloc, idx, cache)*right**xloc*numpy.log(right)

    def _cdf(self, xloc, idx, left, right, cache):
        return left._get_fwd(right.item(0)**xloc, idx, cache)

    def _ppf(self, uloc, idx, left, right, cache):
        return numpy.log(left._get_inv(uloc, idx, cache))/numpy.log(right)

    def _mom(self, kloc, left, right, cache):
        raise chaospy.UnsupportedFeature("%s: Analytical moments for logarithm not supported", self)

    def _ttr(self, kloc, idx, left, right, cache):
        raise chaospy.UnsupportedFeature("%s: Analytical TTR for logarithm not supported", self)


class Log(Logn):
    """
    Logarithm with base Euler's constant.

    Args:
        dist (Distribution):
            Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log(chaospy.Uniform(1, 2))
        >>> distribution
        Log(Uniform(lower=1, upper=2))

    """

    def __init__(self, dist):
        super(Log, self).__init__(dist=dist, base=numpy.e)
        self._repr_args = [dist]


class Log10(Logn):
    """
    Logarithm with base 10.

    Args:
        dist (Distribution): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log10(chaospy.Uniform(1, 2))
        >>> distribution
        Log10(Uniform(lower=1, upper=2))

    """

    def __init__(self, dist):
        super(Log10, self).__init__(dist=dist, base=10)
        self._repr_args = [dist]



"""Tukey-lambda distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class tukey_lambda(SimpleDistribution):
    """Tukey-lambda distribution."""

    def __init__(self, lam):
        super(tukey_lambda, self).__init__(dict(lam=lam))

    def _upper(self, lam):
        return 1./numpy.clip(lam, a_min=0.03, a_max=None)

    def _lower(self, lam):
        return -1./numpy.clip(lam, a_min=0.03, a_max=None)

    def _pdf(self, x, lam):
        lam = numpy.zeros(x.shape) + lam
        output = numpy.zeros(x.shape)
        indices = (lam <= 0) | (numpy.abs(x)*lam < 1)
        lam = lam[indices]
        Fx = special.tklmbda(x[indices], lam)
        Px = 1/(Fx**(lam-1.0) + ((1-Fx))**(lam-1.0))
        output[indices] = Px
        return output

    def _cdf(self, x, lam):
        return special.tklmbda(x, lam)

    def _ppf(self, q, lam):
        output = numpy.zeros(q.shape)
        lam = numpy.broadcast_to(lam, q.shape)
        indices = lam != 0
        q_ = q[indices]
        lam_ = lam[indices]
        output[indices] = (q_**lam_ - (1-q_)**lam_)/lam_
        q_ = q[~indices]
        output[~indices] = numpy.log(q_/(1-q_))
        return output


class TukeyLambda(ShiftScaleDistribution):
    """
    Tukey-lambda distribution.

    Args:
        lam (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.TukeyLambda(1.5)
        >>> distribution
        TukeyLambda(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-0.667, -0.417, -0.141,  0.141,  0.417,  0.667])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.745, 0.711, 0.711, 0.745, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.216, -0.529,  0.61 , -0.025])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(TukeyLambda, self).__init__(
            dist=tukey_lambda(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Baseclass for all conditional distributions."""
import numpy
import chaospy

from .distribution import Distribution


class ItemDistribution(Distribution):
    """Baseclass for an index of multivariate distribution."""

    def __init__(self, index, parent):
        super(ItemDistribution, self).__init__(
            parameters=dict(index=index, parent=parent),
            dependencies=[parent._dependencies[index].copy()],
            rotation=[0],
            repr_args=[index, parent],
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        """Get distribution parameters."""
        assert idx == 0 or idx is None, "Indexes only have a single component"
        if idx == 0:
            idx = int(self._parameters["index"])
        parent = self._parameters["parent"]
        parameters = parent.get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        return dict(parent=parent, parameters=parameters)

    def __repr__(self):
        return "ItemDistribution(%(index)d, %(parent)s)" % self._parameters

    def _lower(self, parent, parameters):
        return parent._lower(**parameters)

    def _upper(self, parent, parameters):
        return parent._upper(**parameters)

    def _ppf(self, xloc, parent, parameters):
        return parent._ppf(xloc, **parameters)

    def _cdf(self, xloc, parent, parameters):
        return parent._cdf(xloc, **parameters)

    def _pdf(self, xloc, parent, parameters):
        return parent._pdf(xloc, **parameters)

    def _mom(self, kloc, parent, parameters):
        idx = int(self._parameters["index"])
        kloc = kloc*numpy.eye(len(parent), dtype=int)[idx]
        return parent._mom(kloc, **parameters)

    def _ttr(self, kloc, parent, parameters):
        raise chaospy.StochasticallyDependentError("TTR not supported")

    def _cache(self, idx, cache):
        if idx is None:
            return self
        assert idx == 0
        idx = int(self._parameters["index"])
        parent = self._parameters["parent"]
        return parent._get_cache(idx, cache)



"""
Create `antithetic variates`_ from variables on the unit hyper-cube.

In statistics, the antithetic variates method is a variance reduction technique
used in Monte Carlo methods. Considering that the error reduction in the
simulated signal (using Monte Carlo methods) has a square root convergence,
a very large number of sample paths is required to obtain an accurate result.
The antithetic variates method reduces the variance of the simulation results.

Antithetic variate can be accessed as a flag ``antithetic`` in the method
``Distribution.sample`` It can either be set to ``True``, for activation, or as an
array of boolean values, which implies it will be used as the flag ``axes`` in
the examples below.

Example usage
-------------

Creating antithetic variates can be done directly from each distribution by
using the ``antithetic`` flag::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> samples = distribution.sample(6, antithetic=True)

Antithetic variates contains compliment values of itself::

    >>> samples.round(4)
    array([0.6536, 0.3464, 0.115 , 0.885 , 0.9503, 0.0497])
    >>> 1-samples.round(4)
    array([0.3464, 0.6536, 0.885 , 0.115 , 0.0497, 0.9503])

Antithetic variates can also be used in multiple dimensions::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> samples = distribution.sample(6, antithetic=True)
    >>> samples.round(4)
    array([[0.0407, 0.9593, 0.0407, 0.9593, 0.3972, 0.6028],
           [0.8417, 0.8417, 0.1583, 0.1583, 0.2071, 0.2071]])
    >>> 1-samples.round(4)
    array([[0.9593, 0.0407, 0.9593, 0.0407, 0.6028, 0.3972],
           [0.1583, 0.1583, 0.8417, 0.8417, 0.7929, 0.7929]])

Lastly, it is also possible to select which axes should be included when
applying the variate by passing a bool array. For axes that are "false", the
value is frozen in place::

    >>> samples = distribution.sample(6, antithetic=[True, False])
    >>> samples.round(4)
    array([[0.3922, 0.6078, 0.1823, 0.8177, 0.7435, 0.2565],
           [0.0696, 0.0696, 0.8853, 0.8853, 0.9526, 0.9526]])
    >>> 1-samples.round(4)
    array([[0.6078, 0.3922, 0.8177, 0.1823, 0.2565, 0.7435],
           [0.9304, 0.9304, 0.1147, 0.1147, 0.0474, 0.0474]])
    >>> samples = distribution.sample(6, antithetic=[False, True])
    >>> samples.round(4)
    array([[0.9311, 0.9311, 0.4154, 0.4154, 0.029 , 0.029 ],
           [0.982 , 0.018 , 0.3396, 0.6604, 0.7067, 0.2933]])
    >>> 1-samples.round(4)
    array([[0.0689, 0.0689, 0.5846, 0.5846, 0.971 , 0.971 ],
           [0.018 , 0.982 , 0.6604, 0.3396, 0.2933, 0.7067]])

.. _antithetic variates: https://en.wikipedia.org/wiki/Antithetic_variates
"""
import numpy


def create_antithetic_variates(samples, axes=()):
    """
    Generate antithetic variables.

    Args:
        samples (numpy.ndarray):
            The samples, assumed to be on the [0, 1]^D hyper-cube, to be
            reflected.
        axes (tuple):
            Boolean array of which axes to reflect. If This to limit the number
            of points created in higher dimensions by reflecting all axes at
            once.

    Returns (numpy.ndarray):
        Same as ``samples``, but with samples internally reflected. roughly
        equivalent to ``numpy.vstack([samples, 1-samples])`` in one dimensions.
    """
    samples = numpy.asfarray(samples)
    assert numpy.all(samples <= 1) and numpy.all(samples >= 0), (
        "all samples assumed on interval [0, 1].")
    if len(samples.shape) == 1:
        samples = samples.reshape(1, -1)
    inverse_samples = 1-samples
    dims = len(samples)

    if not len(axes):
        axes = (True,)
    axes = numpy.asarray(axes, dtype=bool).flatten()

    indices = {tuple(axes*idx) for idx in numpy.ndindex((2,)*dims)}
    indices = sorted(indices, reverse=True)
    indices = sorted(indices, key=lambda idx: sum(idx))
    out = [numpy.where(idx, inverse_samples.T, samples.T).T for idx in indices]
    out = numpy.dstack(out).reshape(dims, -1)
    return out



# pylint: disable=wildcard-import
"""
One of the backbone of any uncertainty quantification is a collection of
probability distributions, and ``chaospy`` is no exception. For example, to
create a Gaussian random variable::

    >>> distribution = chaospy.Normal(mu=2, sigma=2)

The syntax for using distribution is here very similar to the syntax used in
``scipy.dist``. For example, to create values from the *probability density
function*::

    >>> t = numpy.linspace(-3, 3, 9)
    >>> distribution.pdf(t).round(3)
    array([0.009, 0.021, 0.043, 0.078, 0.121, 0.164, 0.193, 0.198, 0.176])

Similarly to create values from the *cumulative distribution function*::

    >>> distribution.cdf(t).round(3)
    array([0.006, 0.017, 0.04 , 0.085, 0.159, 0.266, 0.401, 0.55 , 0.691])

To be able to perform any Monte Carlo method, each distribution contains
*random number generator*::

    >>> distribution.sample(6).round(4)
    array([ 2.7901, -0.4006,  5.2952,  1.9107,  4.2763,  0.4033])

The sample scheme also has a few advanced options. For example, to create
low-discrepancy Hammersley sequences samples combined with antithetic variates::

    >>> distribution.sample(size=6, rule="halton", antithetic=True).round(4)
    array([ 3.349 ,  0.651 , -0.3007,  4.3007,  2.6373,  1.3627])

For a full overview of these options, see :ref:`sampling`

In addition a function for extracting raw-statistical moments is available::

    >>> distribution.mom([0, 1, 2, 3, 4])
    array([  1.,   2.,   8.,  32., 160.])

Note that these are raw moments, not the classical moments with adjustments.
For example, the variance is defined as follows::

    >>> distribution.mom(2) - distribution.mom(1)**2
    4.0

However, if the adjusted moments are of interest, the can be retrieved using
the tools described in :ref:`descriptives`::

    >>> chaospy.Var(distribution)
    array(4.)

Random Seed
-----------

To be able to reproduce results it is possible to fix the random seed in
``chaospy``. For simplicity, The library respect ``numpy.random.seed``. E.g.::

    >>> numpy.random.seed(1234)
    >>> distribution.sample(5).round(4)
    array([0.2554, 2.622 , 1.6865, 3.5808, 3.5442])
    >>> numpy.random.seed(1234)
    >>> distribution.sample(5).round(4)
    array([0.2554, 2.622 , 1.6865, 3.5808, 3.5442])
    >>> distribution.sample(5).round(4)
    array([0.79  , 0.8132, 3.6967, 5.459 , 4.3098])

"""
from .baseclass import *
from .sampler import *
from .collection import *
from .copulas import *
from .operators import *
from .constructor import construct
from .approximation import *
from .kernel import GaussianKDE

from . import (
    baseclass, sampler, approximation,
    copulas, collection, operators, constructor
)



"""Generalized extreme value distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class generalized_extreme(SimpleDistribution):
    """Generalized extreme value distribution."""

    def __init__(self, c=1):
        super(generalized_extreme, self).__init__(dict(c=c))

    def _lower(self, c):
        out = numpy.where(c == 0, -3.5, -numpy.expm1(c*3.5)/c)
        out = numpy.where(c < 0, 1./c, out)
        return out

    def _upper(self, c):
        out = numpy.where(c == 0, 25, -numpy.expm1(-c*25)/c)
        out = numpy.where(c > 0, 1./c, out)
        return out

    def _pdf(self, x, c):
        cx = c*x
        logex2 = numpy.where(c == 0, 0., numpy.log1p(-cx))
        logpex2 = numpy.where(c == 0, -x, logex2/c)
        pex2 = numpy.exp(logpex2)
        logpdf = numpy.where((cx==1) | (cx==-numpy.inf), -numpy.inf, -pex2+logpex2-logex2)
        numpy.putmask(logpdf,(c==1) & (x==1),0.0)
        return numpy.exp(logpdf)

    def _cdf(self, x, c):
        loglogcdf = numpy.where(c == 0, -x, numpy.log1p(-c*x)/c)
        return numpy.exp(-numpy.exp(loglogcdf))

    def _ppf(self, q, c):
        x = -numpy.log(-numpy.log(q))
        return numpy.where(c == 0, x, -numpy.expm1(-c*x)/c)


class GeneralizedExtreme(ShiftScaleDistribution):
    """
    Generalized extreme value distribution.

    Also known as the Fisher-Tippett distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        loc (float, Distribution):
            Location parameter

    Example:
        >>> distribution = chaospy.GeneralizedExtreme(0.5)
        >>> distribution
        GeneralizedExtreme(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-9.509, -0.537,  0.086,  0.571,  1.055,  2.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.254, 0.383, 0.429, 0.378, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.696, -0.941,  1.548,  0.292])

    """

    def __init__(self, shape=0, scale=1, shift=0):
        super(GeneralizedExtreme, self).__init__(
            dist=generalized_extreme(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""
Constructing custom probability distributions is done by subclassing the
distribution :class:`~chaospy.distributions.baseclass.simple.SimpleDistribution`::

    >>> class Uniform(chaospy.SimpleDistribution):
    ...     def __init__(self, lo=0, up=1):
    ...         '''Initializer.'''
    ...         super().__init__(parameters=dict(lo=lo, up=up))
    ...     def _cdf(self, x_data, lo, up):
    ...         '''Cumulative distribution function.'''
    ...         return (x_data-lo)/(up-lo)
    ...     def _lower(self, lo, up):
    ...         '''Lower bound.'''
    ...         return lo
    ...     def _upper(self, lo, up):
    ...         '''Upper bound.'''
    ...         return up
    ...     def _pdf(self, x_data, lo, up):
    ...         '''Probability density function.'''
    ...         return 1./(up-lo)
    ...     def _ppf(self, q_data, lo, up):
    ...         '''Point percentile function.'''
    ...         return q_data*(up-lo)+lo

Custom distributions needs to be initialized and wrapped. For example::

    >>> dist = Uniform(-3, 3)
    >>> dist.fwd([-3, 0, 3])  # Forward Rosenblatt transformation
    array([0. , 0.5, 1. ])

Here the method ``_cdf`` is an absolute requirement. In addition, either
``_ppf``, or the couple ``_lower`` and ``_upper`` should be provided. The
others are not required, but may increase speed and or accuracy of
calculations. In addition to the once listed, it is also
possible to define the following methods:

``_mom``
    Method for creating raw statistical moments, used by the ``mom`` method.
``_ttr``
    Method for creating coefficients from three terms recurrence method, used to
    perform "analytical" Stiltjes' method.
"""
import numpy
import chaospy

from .distribution import Distribution


class SimpleDistribution(Distribution):
    """Distribution for the core probability distribution."""

    def __init__(
            self,
            parameters=None,
            rotation=None,
            exclusion=None,
            repr_args=None,
    ):
        """
        Args:
            parameters (Optional[Distribution[str, Union[ndarray, Distribution]]]):
                Collection of model parameters.
            rotation (None, Sequence[int], Sequence[Sequence[bool]]):
                The order of which to resolve conditionals. Either as
                a sequence of column rotations, or as a permutation
                matrix. Defaults to `range(len(distribution))` which
                is the same as `p(x0), p(x1|x0), p(x2|x0,x1), ...`.
            repr_args (Optional[Sequence[str]]):
                Positional arguments to place in the object string
                representation. The repr output will then be:
                `<class name>(<arg1>, <arg2>, ...)`.

        """
        if parameters is None:
            parameters = {}
        dependencies, parameters, rotation = chaospy.declare_dependencies(
            distribution=self,
            parameters=parameters,
            rotation=rotation,
        )
        super(SimpleDistribution, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            rotation=rotation,
            exclusion=exclusion,
            repr_args=repr_args,
        )


    def get_parameters(self, idx, cache, assert_numerical=True):
        """
        Get distribution parameters.

        Uses the cache to replace parameters that are distribution with cached
        values.

        Args:
            cache (Dict[Distribution, numpy.ndarray]):
                Collection of already calculated values.

        Returns:
            (Dict[str, numpy.ndarray]):
                Collection of parameters. Probability distributions
                are replaced with cached values.

        Raise:
            UnsupportedFeature:
                If a parameter is a probability distribution
                without cache, it means the dependency is
                unresolved.

        """
        parameters = super(SimpleDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        del parameters["cache"]
        del parameters["idx"]
        for key, value in parameters.items():
            if isinstance(value, Distribution):
                value = value._get_cache(idx, cache, get=0)
                assert not assert_numerical or not isinstance(value, Distribution)
                parameters[key] = value
            if idx is not None and len(value) > 1:
                parameters[key] = value[idx]
        return parameters

    def _mom(self, kloc, **kwargs):
        """Default moment generator, throws error."""
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical raw moments." % self)

    def _ttr(self, kloc, **kwargs):
        """Default TTR generator, throws error."""
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical recurrence coefficients." % self)



"""
Newton-Cotes quadrature, are a group of formulas for numerical integration
based on evaluating the integrand at equally spaced points.

Example usage
-------------

Generate Newton-Cotes quadrature rules::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in range(5):
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="newton_cotes")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0. 1.]] [0.5 0.5]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.    0.333 0.667 1.   ]] [0.125 0.375 0.375 0.125]
    4 [[0.   0.25 0.5  0.75 1.  ]] [0.078 0.356 0.133 0.356 0.078]

The first few orders with exponential growth rule::

    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="newton_cotes", growth=True)
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    2 [[0.   0.25 0.5  0.75 1.  ]] [0.078 0.356 0.133 0.356 0.078]
    3 [[0.    0.125 0.25  0.375 0.5   0.625 0.75  0.875 1.   ]]
       [ 0.035  0.208 -0.033  0.37  -0.16   0.37  -0.033  0.208  0.035]

Applying Smolyak sparse grid on Newton-Cotes::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="newton_cotes",
    ...     growth=True, sparse=True)
    >>> abscissas.round(3)
    array([[0.  , 0.  , 0.  , 0.25, 0.5 , 0.5 , 0.5 , 0.5 , 0.5 , 0.75, 1.  ,
            1.  , 1.  ],
           [0.  , 0.5 , 1.  , 0.5 , 0.  , 0.25, 0.5 , 0.75, 1.  , 0.5 , 0.  ,
            0.5 , 1.  ]])
    >>> weights.round(3)
    array([ 0.028,  0.022,  0.028,  0.356,  0.022,  0.356, -0.622,  0.356,
            0.022,  0.356,  0.028,  0.022,  0.028])
"""
from __future__ import division
try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache
import numpy
from scipy.integrate import newton_cotes

from .combine import combine_quadrature


def quad_newton_cotes(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate the abscissas and weights in Newton-Cotes quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Examples:
        >>> abscissas, weights = quad_newton_cotes(4)
        >>> abscissas.round(4)
        array([[0.  , 0.25, 0.5 , 0.75, 1.  ]])
        >>> weights.round(4)
        array([0.0778, 0.3556, 0.1333, 0.3556, 0.0778])
        >>> abscissas, weights = quad_newton_cotes(4, segments=2)
        >>> abscissas.round(4)
        array([[0.  , 0.25, 0.5 , 0.75, 1.  ]])
        >>> weights.round(4)
        array([0.1667, 0.6667, 0.3333, 0.6667, 0.1667])
    """
    from ..distributions.baseclass import Distribution
    if isinstance(domain, Distribution):
        abscissas, weights = quad_newton_cotes(
            order, (domain.lower, domain.upper), growth, segments)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = domain
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()
    dim = max(lower.size, upper.size, order.size)

    order = order*numpy.ones(dim, dtype=int)
    lower = lower*numpy.ones(dim)
    upper = upper*numpy.ones(dim)
    segments = segments*numpy.ones(dim, dtype=int)

    results = [_newton_cotes(*args, growth=growth)
               for args in zip(order, lower, upper, segments)]
    abscissas = [args[0] for args in results]
    weights = [args[1] for args in results]
    return combine_quadrature(abscissas, weights)


@lru_cache(None)
def _newton_cotes(order, lower, upper, segments=1, growth=False):
    """Backend for Newton-Cotes quadrature rule."""
    if order == 0:
        return numpy.array([0.5*(lower+upper)]), numpy.ones(1)
    order = 2**order if growth else order

    if segments != 1 and order > 2:
        if not segments:
            segments = int(numpy.sqrt(order))
        assert segments < order, "few samples to distribute than intervals"
        abscissas = []
        weights = []

        nodes = numpy.linspace(0, 1, segments+1)
        for lower, upper in zip(nodes[:-1], nodes[1:]):

            abscissa, weight = _newton_cotes(order//segments, lower, upper)
            weight = weight*(upper-lower)
            if abscissas:
                weights[-1] += weight[0]
                abscissa = abscissa[1:]
                weight = weight[1:]
            abscissas.extend(abscissa)
            weights.extend(weight)

        assert len(abscissas) == order+1
        assert len(weights) == order+1
        return numpy.array(abscissas), numpy.array(weights)

    return (
        numpy.linspace(lower, upper, order+1),
        newton_cotes(order)[0]/(upper-lower)/order,
    )
